#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@Author: lockerzhang
@LastEditors: lockerzhang
@Description: 小程序层面的操作(包含 web 和 Native)
@Date: 2019-03-11 14:41:43
@LastEditTime: 2019-06-05 16:30:44
"""
import typing
import types
from enum import Enum

from minium.utils.utils import AsyncCondition
from .page import Page
from .minium_object import MiniumObject, RetryableApi
from ...framework.exception import *
from ...utils.injectjs import getInjectJsCode, JsMode
from ...utils.platforms import *
from ...utils.emitter import ee
from ...utils.eventloop import event_loop
from ...utils.utils import get_result, WaitTimeoutError, retry, catch, urlencode
from .callback import AsyncCallback, Callback
import os
import json
import threading
import base64
import io
import time
import datetime
import copy


cur_path = os.path.dirname(os.path.realpath(__file__))  # source_path是存放资源文件的路径
conf_path = os.path.join(os.path.dirname(cur_path), "conf/iOS_conf")


class MockNetworkType(Enum):
    ALWAYS = 0
    ONCE = 1


class MockCloudCallType(Enum):
    ALWAYS = 0
    ONCE = 1


class RouteCommand:
    def __init__(
        self, open_type, path, route_done_lock=None, default_route_result=None
    ) -> None:
        """页面跳转指令

        :param str open_type: 跳转操作: navigateTo/redirectTo/reLaunch/switchTab
        :param str path: 路径: /pages/index/index?a=b&c=1
        :param AsyncCondition route_done_lock: onAppRouteDone回调释放的异步锁, defaults to None(不需要等待route done)
        :param bool default_route_result: _description_, defaults to None
        """
        self.open_type = open_type
        self.path = path
        self._open_id = None  # 调用异步接口返回的请求id

        self._async_msg_lock: AsyncCondition = route_done_lock
        self._open_callback = AsyncCallback(event_loop)
        self._route_changed = None
        self._error: Exception = None

    @property
    def open_id(self):
        return self._open_id

    @open_id.setter
    def open_id(self, value):
        if value == self._open_id:
            return
        self._open_id = value
        ee.once(self._open_id, self.open_callback)

    def on_connect_state_change(self, value: bool, error: Exception):
        if not value:  # break
            self._error = error

    async def _wait_msg_lock(self, timeout=None):
        await self._async_msg_lock.acquire()
        ret = await self._async_msg_lock.wait(timeout=timeout)
        self._async_msg_lock.release()
        return ret

    async def _wait_route_changed_async(self, timeout=None) -> bool:
        if not timeout:
            timeout = 5

        ret = await self._wait_msg_lock(timeout)
        return ret

    def __enter__(self):
        ee.on("connect_state_change", self.on_connect_state_change)
        return self

    def __exit__(self, *args):
        ee.remove_listener("connect_state_change", self.on_connect_state_change)

    def listen_route_change(self, timeout):
        if not self._async_msg_lock:
            return
        self._route_changed = event_loop.run_coroutine(
            self._wait_route_changed_async(timeout + 5)
        )
        return self._route_changed

    def open_callback(self, *args):
        return self._open_callback.callback(*args)

    def cancel_open_callback(self):
        if self._open_id:
            ee.remove_listener(self._open_id, self.open_callback)
        return self._open_callback.cancel()

    def get_open_result(self, timeout):
        """获取打开页面操作结果"""
        try:
            if self._open_callback.wait_called(timeout):
                return self._open_callback.get_result()
            raise MiniTimeoutError(
                f"[{self._open_id}][{self.open_type}] receive from remote timeout"
            )
        except MiniClientOfflineError as co:
            raise MiniTimeoutCauseByClientOffline(str(co)) from co
        except MiniConnectionClosedError as cc:
            raise MiniTimeoutCauseByConnectionBreakError(str(cc)) from cc

    def get_route_result(self, timeout):
        if not self._async_msg_lock:
            # 不需要等待则认为已经route done
            return True
        if self._error:
            raise self._error
        try:
            if get_result(self._route_changed, timeout):
                return True
            return False
        except WaitTimeoutError as wt:
            if self._error:
                if isinstance(self._error, MiniClientOfflineError):
                    raise MiniTimeoutCauseByClientOffline(str(self._error)) from wt
                if isinstance(self._error, MiniConnectionClosedError):
                    raise MiniTimeoutCauseByConnectionBreakError(
                        str(self._error)
                    ) from wt
                raise self._error from wt
            if self.open_type == "switchTab":
                # switchTab方法可能没有on app route done
                return False
            raise wt
        except MiniClientOfflineError as co:
            raise MiniTimeoutCauseByClientOffline(str(co)) from co
        except MiniConnectionClosedError as cc:
            raise MiniTimeoutCauseByConnectionBreakError(str(cc)) from cc


class App(MiniumObject, metaclass=RetryableApi):
    __RETRY_API__ = [
        "enable_log",
        "screen_shot",
        "expose_function",
        "get_all_pages_path",
        "get_current_page",
        # "get_page_stack",  # _page_stack方法重试
        "wait_util",
        "hook_wx_method",
        "release_hook_wx_method",
        "hook_current_page_method",
        "restore_wx_method",
        "restore_request",
        "mock_request",
        # "mock_request_once",  #如果重试, 会影响once的逻辑
        "restore_call_function",
        "mock_call_function",
        # "mock_call_function_once",  #如果重试, 会影响once的逻辑
        "restore_call_container",
        "mock_call_container",
        # "mock_call_container_once",  #如果重试, 会影响once的逻辑
        "edit_editor_text",
        "reflesh_mocked_images",
        "mock_choose_images",  # 重试可能遇上前面已经mock的情况, 但已经兼容
        "mock_choose_image_with_name",
        # protect(common method)
        "_mock_wx_method",  # 各类mock方法都使用到
        "_page_stack",
        # private
        "__get_pages_config",
    ]

    def __init__(self, connection, relaunch=False, native=None, platform="ide"):
        """
        创建一个 APP 实例
        :param connection:
        """
        super().__init__()
        self.connection = connection
        self.native = native
        self.platform = platform
        self.js_mode: JsMode = (
            JsMode.JUST_ES5 if platform in [OS_ANDROID, OS_IOS] else JsMode.ALL
        )  # 真机调试2.0环境下只支持es5语法
        self._msg_lock = threading.Condition()
        self._async_msg_lock = AsyncCondition(loop=event_loop)
        self._is_log_enable = False
        self._main_page_path = None
        self.route_return_page = None
        self.route_return_page_update_time = None  # 监听到页面变化后由于是异步通信, 有可能出现两个相近信息重复顺序调乱
        self._appid = None
        self._current_page = None  # 只有调用 get_current_page 和 监听到 _on_route_changed 回调时更新
        self._mocked_images = None
        self._mocked_images_dir = None
        self._mocked_images_data = {}
        # init methods
        self._route_change_listener()
        self._request_stack_listener()
        self.__is_injected = self._evaluate_js(
            "checkInject"
        )  # 是否已经注入过了, 注入过的环境不需要重复注入, 防止出现多次回调等预期之外的情况
        self.pages = []
        self.tabbar_pages = {}
        self.__get_pages_config()
        if not self.__is_injected and self.js_mode is JsMode.JUST_ES5:  # 注入一些垫片
            self._evaluate_js("helpers")
        if relaunch:
            self.relaunch(self.main_page_path.path, self.main_page_path.query)

    @property
    def app_id(self):
        if not self._appid:
            self._appid = self._get_account_info_sync().result.result.miniProgram.appId
        return self._appid

    @property
    def current_page(self):
        return self._current_page or self.get_current_page()

    @property
    def is_injected(self):
        """
        是否已经注入过, 如果没有, 则设置并进行注入操作
        """
        if self.__is_injected:
            return True
        self.__is_injected = True
        return False

    @property
    def main_page_path(self):
        if self._main_page_path is None:
            options = self._get_launch_options_sync().result.result
            if options.path == "":
                raise MiniLaunchError("get launch options failed")
            self._main_page_path = Page(None, options.path, options.query, None)
        return self._main_page_path

    def _evaluate_js(
        self,
        filename,
        args=None,
        sync=True,
        default=None,
        code_format_info=None,
        **kwargs,
    ):
        """
        重写, 默认使用self.js_mode参数
        """
        return super(App, self)._evaluate_js(
            filename, args, sync, default, code_format_info, self.js_mode, **kwargs
        )

    def enable_log(self):
        """
        打开日志
        每次调用, 都会新增一个回调, 最好记录一下已经调用过
        """
        if not self._is_log_enable:
            self.connection.send("App.enableLog")
            self._is_log_enable = True

    @catch(MiniClientOfflineError)
    def exit(self):
        """
        退出小程序
        :return: None
        """
        setattr(self.connection, "_is_close_app_by_cmd", True)
        self.connection.send_async("App.exit", ignore_response=True)

    def get_account_info_sync(self):
        """
        获取账号信息
        :return:
        """
        return self._get_account_info_sync()

    def screen_shot(self, save_path=None, format="raw", use_native=False):
        """
        截图, 仅能截取 webview 部分, 原生控件不能截取到
        :param save_path:
        :param format: raw or pillow
        :param use_native: use native interface
        :return:
        """
        if self.platform != "ide":
            # 小程序不是运行在开发者工具中, 不支持`App.captureScreenshot`
            if self.native:
                # 使用native接口实现截图
                _save_path = save_path or "%s.png" % datetime.datetime.now().strftime(
                    "%H%M%S%f"
                )
                self.native.screen_shot(_save_path)
                if not os.path.isfile(_save_path):
                    self.logger.error("%s.screen_shot fail" % self.native.__class__)
                    return None
                with open(_save_path, "rb") as fd:
                    raw_value = fd.read()
                if not save_path:
                    os.remove(_save_path)
            else:
                self.logger.error("native instance not exists, please check")
        else:
            try:
                b64_data = (
                    self.connection.send("App.captureScreenshot")
                    .get("result")
                    .get("data")
                )
            except MiniAppError as e:
                self.logger.error(e)
                return None
            else:
                raw_value = base64.b64decode(b64_data)
                png_header = b"\x89PNG\r\n\x1a\n"
                if not raw_value.startswith(png_header) and save_path:
                    self.logger.error("screenshot png format error")

                if save_path:
                    with open(save_path, "wb") as f:
                        f.write(raw_value)

        if format == "raw":
            return raw_value
        elif format == "pillow":
            from PIL import Image

            buff = io.BytesIO(raw_value)
            return Image.open(buff)
        else:
            self.logger.warning(f"unknown format:{format} for screenshot")
            return raw_value

    def expose_function(self, name, binding_function):
        """
        在 AppService 全局暴露方法, 供小程序侧调用测试脚本中的方法。
        :param name:
        :param binding_function:
        :return:
        """
        self._expose_function(name, binding_function)

    def add_observer(self, event, callback):
        """
        监听小程序的事件
        :param event: 需要监听的事件
        :param callback: 收到事件后的回调函数
        :return:
        """
        if not callable(callback):
            raise Exception("the callback is not a function")

        self.connection.register(event, callback)

    def remove_observer(self, event, callback=None):
        """
        移除小程序事件监听
        :param event: 监听的事件
        :param callback: 指定的监听函数, 不传则移除该事件所有监听函数
        :return:
        """
        self.connection.remove(event, callback)

    def on_exception_thrown(self, func):
        """
        JS 错误回调
        :param func: 回调函数
        :return:
        """
        self.connection.register("App.exceptionThrown", func)

    def call_wx_method(self, method, args=None, plugin_appid="auto"):
        """
        调用 wx 的方法
        :param method:
        :param args:
        :param plugin_appid: 调用插件[${appid}]中的方法, auto时自动获取当前页面是否插件页
        :return:
        """
        if plugin_appid == "auto":
            plugin_appid = (self._current_page or self.get_current_page()).plugin_appid
        return self._call_wx_method(method=method, args=args, plugin_appid=plugin_appid)

    def call_wx_method_async(
        self, method, args=None, plugin_appid="auto", ignore_response=False
    ):
        """
        调用 wx 的方法, 返回异步message id
        :param method:
        :param args:
        :param plugin_appid: 调用插件[${appid}]中的方法, auto时自动获取当前页面是否插件页
        :return: str(message_id)
        """
        if plugin_appid == "auto":
            plugin_appid = (self._current_page or self.get_current_page()).plugin_appid
        return self._call_wx_method(
            method=method,
            args=args,
            plugin_appid=plugin_appid,
            sync=False,
            ignore_response=ignore_response,
        )

    def get_all_pages_path(self):
        """
        获取所有已配置的页面路径
        :return:
        """
        return self._evaluate_js("getAllPagesPath")

    def __get_pages_config(self):
        """
        获取所有page相关信息
        """
        try:
            ret = self.evaluate(
                """function(){return {"pages": __wxConfig.pages, "tabBar": __wxConfig.tabBar && __wxConfig.tabBar.list} }""",
                sync=True,
            ).result.result
            self.pages = ret.get("pages", [])
            tabbar_list = ret.get("tabBar", None)
            if tabbar_list:
                idx = 0
                for tabbar in tabbar_list:
                    page_path = tabbar["pagePath"].rstrip(".html")
                    if not page_path.startswith("/"):
                        page_path = "/" + page_path
                    self.tabbar_pages[page_path] = tabbar
                    self.tabbar_pages[page_path].update(
                        {"index": idx, "pagePath": page_path.lstrip("/")}
                    )
                    idx += 1
        except MiniAppError:
            return {}
        return ret

    def get_current_page(self) -> Page:
        """
        获取当前顶层页面
        :return: Page 对象
        """

        def get_current_page(cnt=0):
            ret = self.connection.send("App.getCurrentPage")
            if hasattr(ret, "error"):
                raise Exception("Get current page fail, cause: %s" % ret.error)
            if not ret.result and cnt < 3:
                # 返回信息有问题, 没拿到pageId等属性
                return get_current_page(cnt + 1)
            return ret

        ret = get_current_page()
        page = Page(ret.result.pageId, ret.result.path, ret.result.query, self)
        # 记录当前页面
        self._current_page = page
        return page

    def evaluate(self, app_function: str, args=None, sync=False, desc=None):
        """
        向 app Service 层注入代码并执行
        :param app_function:
        :param args:
        :param sync:
        :param desc: 报错描述
        :return:
        """
        return self._evaluate(
            app_function=app_function, args=args, sync=sync, desc=desc
        )

    def get_page_stack(self) -> typing.List[Page]:
        """
        获取当前小程序所有的页面
        :return: Page List
        """
        return self._page_stack()

    # navigate_to不进行重试:
    # 1. 小程序断线没有重试必要, 必须等待小程序重连。
    # 2. ws链接断连在send的时候就已经检查并等待重连。
    # 3. 如果是单纯没有回包, 不能确定指令是否已经生效, 如指令生效并重试, 会增加页面堆栈
    def navigate_to(self, url, params=None, is_wait_url_change=True):
        """
        以导航的方式跳转到指定页面, 但是不能跳到 tabbar 页面。支持相对路径和绝对路径, 小程序中页面栈最多十层
        支持相对路径和绝对路径
        /page/tabBar/API/index: 绝对路径,最前面为/
        tabBar/API/index: 相对路径, 会被拼接在当前页面的路径后面
        :param url:"/page/tabBar/API/index"
        :param params: 页面参数
        :param is_wait_url_change: 是否等待页面变换完成
        :return:Page 对象
        """
        if params:
            url += (
                ("?" + urlencode(params))
                if url.find("?") < 0
                else ("&" + urlencode(params))
            )
        self.logger.info("NavigateTo: %s" % url)
        page = self._change_route_async("navigateTo", url, is_wait_url_change)
        if page != url.split("?")[0]:
            self.logger.warning("NavigateTo(%s) but(%s)" % (url, page.path))
        return page

    # redirect_to因为不产生新的页面堆栈, ws断连引起的回报丢失可以尝试重试
    @retry(2, (MiniTimeoutCauseByConnectionBreakError))
    def redirect_to(self, url, params=None, is_wait_url_change=True):
        """
        关闭当前页面, 重定向到应用内的某个页面。但是不允许跳转到 tabbar 页面
        :param url:"/page/tabBar/API/index"
        :param params: 页面参数
        :param is_wait_url_change: 是否等待页面变换完成
        :return:Page 对象
        """
        if params:
            url += (
                ("?" + urlencode(params))
                if url.find("?") < 0
                else ("&" + urlencode(params))
            )
        self.logger.info("RedirectTo: %s" % url)
        page = self._change_route_async("redirectTo", url, is_wait_url_change)
        if page != url.split("?")[0]:
            self.logger.warning("RedirectTo(%s) but(%s)" % (url, page.path))
        return page

    # relaunch因为不产生新的页面堆栈, ws断连引起的回报丢失可以尝试重试
    @retry(2, (MiniTimeoutCauseByConnectionBreakError))
    def relaunch(self, url, params=None, is_wait_url_change=True):
        """
        关闭所有页面, 打开到应用内的某个页面
        :param url: "/page/tabBar/API/index"
        :return:Page 对象
        """
        if params:
            url += (
                ("?" + urlencode(params))
                if url.find("?") < 0
                else ("&" + urlencode(params))
            )
        self.logger.info("ReLaunch: %s" % url)
        page = self._change_route_async("reLaunch", url, is_wait_url_change)
        if page != url.split("?")[0]:
            self.logger.warning("ReLaunch(%s) but(%s)" % (url, page.path))
        return page

    def navigate_back(self, delta=1):
        """
        关闭当前页面, 返回上一页面或多级页面。
        :param delta: 返回的层数, 如果超出 page stack 最大层数返回首页
        :return:Page 对象
        """
        # self._wait_until_page_is_stable()
        page = self.current_page
        page_stack = self._page_stack()
        if page.page_id == page_stack[0].page_id:
            self.logger.warning("Current page is root, can't navigate back")
            return page
        self.logger.info("NavigateBack from:%s" % page.path)
        self._call_wx_method("navigateBack", [{"delta": delta}])
        if self._route_changed():
            return self.route_return_page
        else:
            self.logger.warning("route has not change, may be navigate back fail")

    # switch_tab因为不产生新的页面堆栈, ws断连引起的回报丢失可以尝试重试
    @retry(2, (MiniTimeoutCauseByConnectionBreakError))
    def switch_tab(
        self, url, is_wait_url_change=True, is_click=False, wait_route_done_time=15
    ):
        """
        跳转到 tabBar 页面, 并关闭其他所有非 tabBar 页面
        :param url: "/page/tabBar/API/index"
        :param is_click: 点击触发帮用户触发一下onTabItemTap
        :return:Page 对象
        """
        page = self._change_route_async(
            "switchTab", url, is_wait_url_change, wait_route_done_time
        )
        if is_click and page.path in self.tabbar_pages:  # is tabbar and is click
            # 需要触发 onTabItemTap
            try:
                page.call_method(
                    "onTabItemTap",
                    {
                        "index": self.tabbar_pages[page.path]["index"],
                        "pagePath": self.tabbar_pages[page.path]["pagePath"],
                        "text": self.tabbar_pages[page.path]["text"],
                    },
                )
            except MiniAppError:
                pass
        if page != url.split("?")[0]:
            self.logger.warning("Switch tab(%s) but(%s)" % (url, page.path))
        return page

    def stop_audits(self, path: str = None):
        """
        停止体验测评
        :path: 存放评测结果html的路径, 默认不存
        :return: dict, 测评结果
        """
        ret = self._stop_audits()
        if path:
            html_result = ret["result"]["report"]
            with open(path, "w", encoding="utf-8") as h_f:
                h_f.write(html_result)
        return json.loads(ret["result"]["data"])

    def get_perf_time(self, entry_types: list = None):
        default_entry_types = ["render", "script", "navigation", "loadPackage"]
        if not entry_types:
            entry_types = default_entry_types
        for item_type in entry_types:
            if item_type not in default_entry_types:
                raise Exception("the entryType is not available")
        self._evaluate_js(
            "startGetPerformance",
            [
                entry_types,
            ],
        )

    def stop_get_perf_time(self):
        time.sleep(2)  # 收集数据的时间推后一点点, 怕收集不完全
        return self._evaluate_js("stopGetPerformance")

    def go_home(self, main_page_path=None):
        """
        跳转到首页
        :param main_page_path: 路径
        :return: Page 对象
        """
        if not main_page_path:
            main_page_path = self.main_page_path
        elif not isinstance(main_page_path, Page):
            main_page_path = Page(None, main_page_path, None, None)

        if main_page_path.path in self.tabbar_pages:
            if (
                self.current_page.path == main_page_path.path
            ):  # switch will not route done
                is_wait_url_change = False
            else:
                is_wait_url_change = True
            # switch_tab有可能没有route done回调, 最多等5s
            page = self.switch_tab(
                main_page_path.path,
                is_wait_url_change=is_wait_url_change,
                wait_route_done_time=5,
            )
        else:
            page = self.relaunch(main_page_path.path, main_page_path.query)
        return page

    def get_async_response(self, msg_id: str, timeout=None) -> None or dict:
        """
        description: 获取异步调用的结果
        param {*} self
        param {str} msg_id 消息ID
        param {int} timeout 超时时间
        return {*}
        """
        if timeout is None:
            return self._get_async_response(msg_id)
        ret = None

        def get():
            nonlocal ret
            ret = self._get_async_response(msg_id)
            return ret is not None

        self._wait(get, timeout)
        return ret

    def wait_for_page(self, page_path=None, max_timeout=10):
        if not self._current_page or self._current_page != page_path:
            # 没有记录到当前页面/当前页面不是目标页面, 等待页面跳转成功的回调
            self._route_changed(max_timeout)
            ret = self.current_page == page_path
            # issue#144 如果等不到预期的页面, 考虑是否因为页面栈满了(自动化流程中, 用户可以直接调用navigate to, 很有可能会满的)
            if not ret:
                if len(self._page_stack()) == 10:
                    self.logger.warning(
                        """
wait for page fail, probably because the page stack is exceeded 10
more infomation see: https://developers.weixin.qq.com/miniprogram/dev/api/route/wx.navigateTo.html#功能描述
"""
                    )
            return ret
        return True

    def wait_util(self, cnt, max_timeout=10):
        """
        {max_timeout}秒内, 剩余没有完成的请求数 <= {cnt}个
        符合上述条件则返回 True, 否则 False
        """
        return self._evaluate_js("waitUtil", [cnt, max_timeout], default=False)

    def _get_launch_options_sync(self):
        return self._call_wx_method("getLaunchOptionsSync")

    def _change_route(self, open_type, path, is_wait_url_change=True) -> Page:
        """
        跳转页面
        call wx method -> call result / on app route done -> on app route done / call result
        "call result" and "on app route done" should wait together
        """
        self.call_wx_method(open_type, [{"url": path}])
        if is_wait_url_change and self._route_changed():
            return self.route_return_page
        else:
            return self.current_page  # todo: 状态有BUG, 超时也认为可用的

    def _change_route_async(
        self, open_type, path, is_wait_url_change=True, wait_route_done_time=15
    ):
        """
        跳转页面
        call wx method -> call result / on app route done -> on app route done / call result
        "call result" and "on app route done" should wait together
        """
        wait_timeout = self.connection.timeout
        with RouteCommand(
            open_type,
            path,
            route_done_lock=self._async_msg_lock if is_wait_url_change else None,
        ) as cmd:
            cmd.listen_route_change(wait_timeout)

            cmd.open_id = self.call_wx_method_async(open_type, [{"url": path}])
            # 理论上RouteCommand中通过ee监听了不会再在这里获取到结果, 以防线程切换过程导致时序问题，这里做一个兜底
            response = self.connection.get_aysnc_msg_return(cmd.open_id)
            if response is None:
                cmd.get_open_result(wait_timeout)
            else:
                cmd.cancel_open_callback()

            if cmd.get_route_result(wait_route_done_time):
                return self.current_page
            else:
                return self.get_current_page()  # 没有等到route done, 直接获取当前页面

    def _on_route_changed(self, message):
        if not message.name == "onAppRouteDone":
            return
        args = message.args
        options = args[0]
        update_time = args[1] if len(args) > 1 else int(time.time() * 1000)
        if (
            self.route_return_page_update_time
            and self.route_return_page_update_time > update_time
        ):
            # 旧信息, 丢弃
            return
        self.route_return_page = Page(
            options.webviewId, options.path, options.query, self
        )
        # 记录当前页面
        self._current_page = self.route_return_page
        self._notify_msg_lock()
        self.logger.info("Route changed, %s" % message)

    def _route_changed(self, timeout=None):
        if not timeout:
            timeout = 5

        self._msg_lock.acquire()
        ret = self._msg_lock.wait(timeout)
        self._msg_lock.release()
        return ret

    def _notify_msg_lock(self):
        async def _notify():
            await self._async_msg_lock.acquire()
            self._async_msg_lock.notify_all()
            self._async_msg_lock.release()

        event_loop.run_coroutine(_notify())
        self._msg_lock.acquire()
        self._msg_lock.notify_all()
        self._msg_lock.release()

    async def _wait_msg_lock(self, timeout=None):
        await self._async_msg_lock.acquire()
        ret = await self._async_msg_lock.wait(timeout=timeout)
        self._async_msg_lock.release()
        return ret

    async def _wait_route_changed_async(self, timeout=None):
        if not timeout:
            timeout = 5

        ret = await self._wait_msg_lock(timeout)
        return ret

    def _route_change_listener(self):
        self._unregister("onAppRouteDone")
        self._expose_function("onAppRouteDone", self._on_route_changed)
        self._evaluate(
            """function () {
    if (!global.__minium_onAppRouteDone) {
        wx.onAppRouteDone(function (options) {
            onAppRouteDone(options, Date.now())
        })
        global.__minium_onAppRouteDone = true
    }
}""",
            sync=True,
        )

    def _request_stack_listener(self):
        return self._evaluate_js("requestStack")

    def _get_account_info_sync(self):
        return self._call_wx_method("getAccountInfoSync")

    def _page_stack(self):
        ret = self.connection.send("App.getPageStack")
        page_stack = []
        for page in ret.result.pageStack:
            page_stack.append(Page(page.pageId, page.path, page.query, self))
        return page_stack

    # ws断连可重试
    @retry(2, (MiniConnectionClosedError))
    def _stop_audits(self):
        ret = self.connection.send("Tool.stopAudits")
        return ret

    def hook_wx_method(
        self,
        method: str,
        before: Callback or types.FunctionType = None,
        after: Callback or types.FunctionType = None,
        callback: Callback or types.FunctionType = None,
    ):
        """
        hook wx 方法
        :param method: 需要 hook 的方法
        :param before: 在需要 hook 的方法之前调用
        :param after: 在需要 hook 的方法之后调用
        :param callback: 在需要 hook 的方法回调之后调用
        :return:
        """
        stages: typing.List[types.FunctionType] = []  # 完成的步骤
        if isinstance(before, Callback):
            before = before.callback
        if isinstance(after, Callback):
            after = after.callback
        if isinstance(callback, Callback):
            callback = callback.callback

        if before and not callable(before):
            self.logger.error(f"wx.{method} hook before method is non-callable")
            return

        if after and not callable(after):
            self.logger.error(f"{method} hook after method is non-callable")
            return

        if callback and not callable(callback):
            self.logger.error(f"{method} hook callback method is non-callable")
            return
        try:

            def super_before(msg):
                self.logger.debug(f"{method} before hook result: {msg['args']}")
                if before:
                    before(msg["args"])

            if before:
                stages.append(super_before)
                self._expose_function(
                    method + "_" + super_before.__name__, super_before
                )

            def super_after(msg):
                self.logger.debug(f"wx.{method} after hook result: {msg['args']}")
                if after:
                    after(msg["args"])

            if after:
                stages.append(super_after)
                self._expose_function(method + "_" + super_after.__name__, super_after)

            def super_callback(msg):
                self.logger.debug(f"wx.{method} callback hook result: {msg['args']}")
                if callback:
                    callback(msg["args"])

            if callback and not method.endswith("Sync"):  # Sync方法没有callback，通过after回调
                stages.append(super_callback)
                self._expose_function(
                    method + "_" + super_callback.__name__, super_callback
                )
            return self._evaluate_js(
                "hookWxMethod",
                code_format_info=dict(
                    method=method,
                    origin=method + "_MiniumOrigin",
                    before=method + "_" + super_before.__name__,
                    after=method + "_" + super_after.__name__,
                    callback=method + "_" + super_callback.__name__,
                ),
            )
        except MiniError:
            for stage in stages:
                self.remove_observer(method + "_" + stage.__name__, stage)
            stages = []
            raise

    def release_hook_wx_method(self, method):
        """
        释放 hook wx 方法
        :param method: 需要释放 hook 的方法
        :return:
        """
        self._evaluate_js(
            "releaseHookWxMethod",
            code_format_info=dict(
                origin=method + "_MiniumOrigin",
                method=method,
                before=method + "_super_before",
                after=method + "_super_after",
                callback=method + "_super_callback",
            ),
        )
        # 移除监听函数
        self.connection.remove(method + "_super_before")
        self.connection.remove(method + "_super_after")
        self.connection.remove(method + "_super_callback")

    def hook_current_page_method(self, method, callback):
        """
        hook 当前页面的方法
        :param method:  方法名
        :param callback:    回调函数
        """

        if isinstance(callback, Callback):
            callback = callback.callback

        def super_callback(msg):
            self.logger.debug(f"Page.{method} call hook result: {msg['args']}")
            if callback:
                callback(msg["args"])

        if callback and not callable(callback):
            self.logger.error(f"Page.{method} hook callback method is non-callable")
            return
        try:
            if callback:
                self._expose_function(
                    "page_hook_" + method + "_" + super_callback.__name__,
                    super_callback,
                )
            self._evaluate_js(
                "hookCurrentPageMethod",
                code_format_info={
                    "method": method,
                    "callback": "page_hook_" + method + "_" + super_callback.__name__,
                },
            )
        except MiniError:
            if callback:
                self.remove_observer(
                    "page_hook_" + method + "_" + super_callback.__name__,
                    super_callback,
                )
            raise

    def release_hook_current_page_method(self, method):
        # 移除监听函数
        self.connection.remove("page_hook_" + method + "_super_callback")

    ###
    # 各类 mock
    ###

    def mock_wx_method(
        self,
        method,
        functionDeclaration: str = None,
        result=None,
        args=None,
        success=True,
        plugin_appid="auto",
    ):
        """
        mock wx method and return result
        :param self:
        :param method:
        :param functionDeclaration:
        :param result:
        :param args:
        :param success:
        :return:
        """
        if plugin_appid == "auto":
            plugin_appid = (self._current_page or self.get_current_page()).plugin_appid
        self._mock_wx_method(
            method=method,
            functionDeclaration=functionDeclaration,
            result=result,
            args=args,
            success=success,
            plugin_appid=plugin_appid,
        )

    def restore_wx_method(self, method, plugin_appid="auto"):
        """
        恢复被 mock 的方法
        :param method: mock的方法
        :param plugin_appid: 插件appid
        :return:
        """
        params = {"method": method}
        if plugin_appid == "auto":
            plugin_appid = (self._current_page or self.get_current_page()).plugin_appid
        if plugin_appid:
            params["pluginId"] = plugin_appid
        self.connection.send("App.mockWxMethod", params)

    def _mock_network(
        self,
        interface: str,
        rule: str or dict,
        success=None,
        fail=None,
        mock_type=MockNetworkType.ALWAYS,
    ):
        if success and fail:
            raise RuntimeError("Can't call back both SUCCESS and FAIL")
        if not (success or fail):
            raise RuntimeError("Must call back either SUCCESS or FAIL")
        if isinstance(rule, (str, bytes)):
            # 默认匹配url
            _rule = {"url": rule}
        else:
            _rule = rule
        if success:
            _rule["success"] = success
        elif fail:
            if isinstance(fail, (str, bytes)):
                fail = {"errMsg": "%s:fail %s" % (interface, fail)}
            _rule["fail"] = fail
        if mock_type == MockNetworkType.ONCE:
            # 保护字段, 保证不冲突
            _rule["_miniMockType"] = MockNetworkType.ONCE.value
        has_mock = self._evaluate_js(
            "addNetworkMockRule", [_rule], code_format_info={"interface": interface}
        )
        if not has_mock:
            retry(self.__class__.__RETRY_CNT__, self.__class__.__RETRY_EXCEPTION__)(
                self._evaluate_js
            )("mockNetwork", code_format_info={"interface": interface})

    def _restore_network(self, interface: str):
        """
        恢复被mock的网络接口
        """
        return self._evaluate_js(
            "cleanNetworkMockRule", code_format_info={"interface": interface}
        )

    def _mock_cloud_call(
        self,
        interface: str,
        rule: dict,
        success=None,
        fail=None,
        mock_type=MockCloudCallType.ALWAYS,
    ):
        if success and fail:
            raise RuntimeError("Can't call back both SUCCESS and FAIL")
        if not (success or fail):
            raise RuntimeError("Must call back either SUCCESS or FAIL")
        if not isinstance(rule, dict):
            raise RuntimeError(
                'Cloud call mock rule must be a dict like {"name": "testCloudFunction"}'
            )
        _rule = copy.deepcopy(rule)
        if success:
            _rule["success"] = success
        elif fail:
            if isinstance(fail, (str, bytes)):
                fail = {"errMsg": "%s:fail %s" % (interface, fail)}
            _rule["fail"] = fail
        if mock_type == MockCloudCallType.ONCE:
            # 保护字段, 保证不冲突
            _rule["_miniMockType"] = MockCloudCallType.ONCE.value
        has_mock = self._evaluate_js("addCloudCallMockRule", [_rule, interface])
        if not has_mock:
            retry(self.__class__.__RETRY_CNT__, self.__class__.__RETRY_EXCEPTION__)(
                self._evaluate_js
            )("mockCloudCall")

    def _restore_cloud_call(self, interface: str):
        """
        恢复被mock的云调用接口
        """
        return self._evaluate_js(
            "cleanCloudCallMockRule",
            [
                interface,
            ],
        )

    def mock_request(self, rule: str or dict, success=None, fail=None):
        """
        mock wx.request, 根据正则mock规则返回mock结果
        """
        return self._mock_network("request", rule, success, fail)

    def mock_request_once(self, rule: str or dict, success=None, fail=None):
        """
        mock wx.request, 根据正则mock规则返回mock结果, 一旦匹配上了, 即废除该rule
        """
        return self._mock_network("request", rule, success, fail, MockNetworkType.ONCE)

    def restore_request(self):
        return self._restore_network("request")

    def mock_call_function(self, rule: dict, success=None, fail=None):
        return self._mock_cloud_call("callFunction", rule, success, fail)

    def mock_call_function_once(self, rule: dict, success=None, fail=None):
        return self._mock_cloud_call(
            "callFunction", rule, success, fail, MockCloudCallType.ONCE
        )

    def mock_call_container(self, rule: dict, success=None, fail=None):
        return self._mock_cloud_call("callContainer", rule, success, fail)

    def mock_call_container_once(self, rule: dict, success=None, fail=None):
        return self._mock_cloud_call(
            "callContainer", rule, success, fail, MockCloudCallType.ONCE
        )

    def restore_call_function(self):
        return self._restore_cloud_call("callFunction")

    def restore_call_container(self):
        return self._restore_cloud_call("callContainer")

    def mock_show_modal(self, answer=True):
        """
        mock 弹窗
        :param answer: 默认点击确定
        :return: None
        """
        self._mock_wx_method(
            "showModal",
            result={
                "cancel": answer,
                "confirm": False if answer else True,
                "errMsg": "showModal:ok",
            },
        )

    def mock_get_location(
        self,
        acc=65,
        horizontal_acc=65,
        vertical_acc=65,
        speed=-1,
        altitude=0,
        latitude=23.12908,
        longitude=113.26436,
    ):
        """
        mock 位置获取
        :param acc: 位置的精确度
        :param horizontal_acc: 水平精度, 单位 m
        :param vertical_acc: 垂直精度, 单位 m(Android 无法获取, 返回 0)
        :param speed: 速度, 单位 m/s
        :param altitude: 高度, 单位 m
        :param latitude: 纬度, 范围为 -90~90, 负数表示南纬
        :param longitude: 经度, 范围为 -180~180, 负数表示西经
        :return:
        """
        self._mock_wx_method(
            "getLocation",
            result={
                "accuracy": acc,
                "altitude": altitude,
                "errMsg": "getLocation:ok",
                "horizontalAccuracy": horizontal_acc,
                "verticalAccuracy": vertical_acc,
                "latitude": latitude,
                "longitude": longitude,
                "speed": speed,
            },
        )

    def mock_show_action_sheet(self, tap_index=0):
        """
        mock 显示操作菜单
        :param tap_index: 用户点击的按钮序号, 从上到下的顺序, 从0开始
        :return:
        """
        self._mock_wx_method(
            "showActionSheet",
            result={"errMsg": "showActionSheet:ok", "tapIndex": tap_index},
        )

    def edit_editor_text(self, editorid, text):
        ret = self._evaluate_js(
            "editEditorText",
            code_format_info={
                "editorid": editorid,
                "text": text,
            },
        )
        self.logger.info(ret)
        return ret

    def reflesh_mocked_images(self, mock_images_dir=None, mock_images=None):
        """
        1. 暴露全局方法用于mock chooseImage
        2. 获取已经有的文件列表
        """
        self._mocked_images_dir = mock_images_dir or None
        self._mocked_images_data = mock_images or {}
        if self._mocked_images is None:
            self._mocked_images = {}  # 需要先调用一次reflesh再进行mock
        ret = self._evaluate_js(
            "mockChooseImage",
        )  # [{name, size}]
        if not ret:
            return
        for item in ret:
            self._mocked_images[item["name"]] = item["size"]

    def mock_choose_image(self, name: str, image_b64data: str):
        """
        mock chooseImage, chooseMedia, takePhoto等选择图片相关接口
        :param name: 文件名
        :param image_b64data: base64格式的图片数据
        :return bool: True for success, False for fail
        """
        return self.mock_choose_images([{"name": name, "b64data": image_b64data}])

    def mock_choose_images(self, items):
        """
        mock chooseImage, chooseMedia, takePhoto等选择图片相关接口
        :param items: [{name, b64data} ... ]
            :name: 文件名
            :b64data: base64格式的图片数据
        :return bool: True for success, False for fail
        """
        if self._mocked_images is None:
            self.reflesh_mocked_images()
        ret = self._evaluate_js(
            "evalMockChooseImage",
            [
                {"imageName": item["name"], "size": self._mocked_images[item["name"]]}
                if item["name"] in self._mocked_images
                else {"imageName": item["name"], "imageData": item["b64data"]}
                for item in items
            ],
        )
        for result in ret:
            self._mocked_images[result["name"]] = result["size"]
        return True

    def mock_choose_image_with_name(self, file_name):
        """
        通过文件名mock上传图片, 被mock的图片需要放在`${mock_images_dir}`或配置在`${mock_images}`键值对中
        :return bool: True for success, False for fail
        """
        if not self._mocked_images_dir and not self._mocked_images_data:
            raise MiniConfigError(
                "can't not use `mock_choose_image`, please config `mock_images_dir` or `mock_images` first"
            )
        if self._mocked_images is None:
            self.reflesh_mocked_images()
        if file_name in self._mocked_images:
            self._evaluate_js(
                "evalMockChooseImage",
                [{"imageName": file_name, "size": self._mocked_images[file_name]}],
            )
            return True
        if file_name in self._mocked_images_data:  # 命中data
            file_content = self._mocked_images_data[file_name]
        else:
            file_path = os.path.join(self._mocked_images_dir, file_name)
            if os.path.isfile(file_path):
                with open(file_path, "rb") as fd:
                    c = fd.read()
                    file_content = base64.b64encode(c).decode("utf8")
            else:
                file_content = ""
        if not file_content:
            return False
        ret = self._evaluate_js(
            "evalMockChooseImage", [{"imageName": file_name, "imageData": file_content}]
        )
        for result in ret:
            self._mocked_images[result["name"]] = result["size"]
        return True



# -*- coding: utf-8 -*-
"""
@author: 'xiazeng'
@created: 2016/12/2 
"""
import re
import time
import datetime
import threading
import sys

if sys.version_info[0] < 3:
    from Queue import Queue, Empty
else:
    from queue import Queue, Empty

import logging

logger = logging.getLogger()


def decode_bytes(bs):
    ds = ['ascii', 'utf-8', 'gbk']
    for d in ds:
        try:
            return bs.decode(d)
        except UnicodeDecodeError:
            pass
    # 可能出现utf-8被截断的场景
    # UnicodeDecodeError: 'utf-8' codec can't decode byte 0xe9 in position 4094: invalid continuation byte
    return bs.decode("utf-8", 'ignore')


class LogRecord(object):
    reg_str = r"(?P<time>(\d{4}-)?\d{2}-\d{2}\s+\d{2}:\d{2}:\d{2}\.\d+)\s+" \
              r"(?P<level>[VDIWEFS])\/" \
              r"(?P<tag>.*(?=\(\s*\d+\s*\)))\(\s*" \
              r"(?P<pid>\d+)\s*\)\s*:\s+" \
              r"(?P<msg>.*)"
    reg = re.compile(reg_str, flags=re.VERBOSE | re.MULTILINE)

    def __init__(self, line):
        self.raw_line = line
        m = LogRecord.reg.match(line)
        if m is None:
            raise RuntimeError("invalided line:" + line)
        self.time = m.group("time")
        self.level = m.group("level")
        self.tag = m.group("tag")
        self.pid = m.group("pid")
        self.msg = m.group("msg")

    @property
    def ts(self):
        t = self.time.split(".")
        if len(self.time.split("-")) != 3:
            time_str = "%s-%s" % (datetime.datetime.now().year, t[0])
        else:
            time_str = int(t[0])
        timetuple = time.strptime(time_str, "%Y-%m-%d %H:%M:%S")
        ts = time.mktime(timetuple) * 1000 + int(t[1])
        return ts

    @classmethod
    def is_valid_line(cls, line):
        m = cls.reg.match(line)
        if not m:
            logger.warning(line)
        return m is not None

    def __repr__(self):
        return "%s %s %s %s" % (self.time, self.level, self.tag, self.msg)


class AsynchronousFileReader(threading.Thread):
    """
    Helper class to implement asynchronous reading of a file
    in a separate thread. Pushes read lines on a queue to
    be consumed in another thread.
    """

    def __init__(self, fd, process, queue):
        assert isinstance(queue, Queue)
        assert callable(fd.readline)
        threading.Thread.__init__(self)
        self._fd = fd
        self._process = process
        self.daemon = True
        self._queue = queue
        self._should_stop = False

    def run(self):
        """The body of the tread: read lines and put them on the queue."""
        count = 0
        next_count = 10000
        start = time.time()
        try:
            while True:
                if self._should_stop:
                    break
                if self._process.poll() is not None:
                    logging.error("process stopped:%s, logcount:%s", self._process.poll(), count)
                    break
                line = self._fd.readline()
                if line:
                    count += 1
                    self._queue.put(line)
                    if count > next_count:
                        t = time.time() - start
                        logger.debug("total:%s, cost:%d, produce rate:%s per second", count, t, int(count / t))
                        next_count += next_count
                    if line == b'read: unexpected EOF!\n':
                        break
        except:
            logger.exception("exception")

        logger.info("read logcat finish: %s", self._should_stop)

    def eof(self):
        """Check whether there is no more content to expect."""
        return not self.is_alive()

    def stop(self):
        self._should_stop = True
        self.join(1)


class JLogCat(threading.Thread):
    def __init__(self, adb):
        threading.Thread.__init__(self)
        self._is_stop = False
        self.adb = adb
        # Launch the asynchronous readers of the process' stdout.
        self.stdout_queue = Queue()
        self._bg_run()
        self.mutex = threading.Lock()
        self.filter_map = {}
        self.filter_values = []
        self.names = []
        self.redirect_file = None

    def redirect_to_path(self, path):
        self.redirect_file = open(path, "a")

    def _bg_run(self, clear_logcat=True):
        if clear_logcat:
            self.adb.run_adb("logcat -c")
        self.process = self.adb.run_adb("logcat  -v time", False)
        logging.info("process_id:%s", self.process.pid)
        stdout_reader = AsynchronousFileReader(self.process.stdout, self.process, self.stdout_queue)
        stdout_reader.start()
        self.stdout_reader = stdout_reader

    def run(self):
        # Check the queues if we received some output (until there is nothing more to get).
        retype = type(re.compile('hello, world'))
        line_count = 0
        start = time.time()
        next_count = 10000

        while not self._is_stop:
            if self.stdout_reader.eof():
                self._bg_run(clear_logcat=False)
            while not self.stdout_queue.empty() and not self._is_stop:
                line = None
                try:
                    line = self.stdout_queue.get(timeout=0.1)
                except Empty:
                    logger.info("empty")
                    pass
                line = decode_bytes(line)
                line_count += 1
                if line_count > next_count:
                    t = time.time() - start
                    logger.debug("total:%s, cost:%d, consume rate:%s per second", line_count, t, int(line_count / t))
                    next_count += next_count
                if line:
                    if self.redirect_file:
                        self.redirect_file.write(line)
                    for fm in self.filter_values:
                        records = fm["records"]
                        for keyword in fm["keywords"]:
                            if isinstance(keyword, str):
                                if keyword in line:
                                    logger.debug("[%s]+logcat, %s", self.ident, line.strip())
                                    records.append(line.strip())
                            elif isinstance(keyword, retype):
                                if keyword.match(line):
                                    logger.debug("[%s]+logcat, %s", self.ident, line.strip())
                                    records.append(line.strip())
        logger.info("finished: %s", self._is_stop)

    def start_record(self, name, *args):
        """
        开始等待，日志
        :param name:标识名称，任意字符串，后面取日志要根据这个标识来区分
        :param *args: 日志匹配的关键字
        :return:
        """
        self.mutex.acquire()
        self.filter_map[name] = {
            "keywords": args,
            "records": []
        }
        self.filter_values = list(self.filter_map.values())
        self.mutex.release()

    def _get_records(self, name):
        self.mutex.acquire()
        if name not in self.filter_map:
            lines = []
        else:
            lines = self.filter_map[name]["records"]
        records = [LogRecord(l) for l in lines if LogRecord.is_valid_line(l)]
        self.mutex.release()
        return records

    def get_lines(self, name):
        return self._get_records(name)

    def wait_records(self, name, count=1, timeout=10):
        s = time.time()
        records = []
        while time.time() - s < timeout:
            records = self._get_records(name)
            if len(records) >= count:
                return records
            time.sleep(0.1)
        return records

    def new_wait_records(self, name, count=1, timeout=20):
        time.sleep(timeout)
        records = self._get_records(name)
        if len(records) < count:
            time.sleep(5)
            records = self._get_records(name)
        return records

    def stop_record(self, name):
        records = self._get_records(name)
        self.mutex.acquire()
        del self.filter_map[name]
        self.filter_values = list(self.filter_map.values())
        self.mutex.release()
        return records

    def stop(self):
        self._is_stop = True
        self.stdout_reader.stop()
        self.join()
        self.adb.kill_pid(self.process.pid)
        if self.redirect_file:
            self.redirect_file.close()
        logger.info("logcat finish")

    def __del__(self):
        self.stop()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Created by xiazeng on 2017/8/28


def is_windows():
    import platform
    pf = platform.platform()
    if "Windows" in pf:
        return True
    else:
        return False


def wait_variable(v):
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@Author: lockerzhang
@LastEditors: lockerzhang
@Description: 元素相关的操作和信息获取
@Date: 2019-03-11 14:42:10
@LastEditTime: 2019-06-06 15:13:32
"""
from __future__ import annotations
import code
from logging import Logger
import typing
from minium.framework.exception import MiniElementNotFoundError, MiniTimeoutCauseByConnectionBreakError
from .minium_object import MiniumObject, RetryableApi
from ...utils.utils import timeout
from ...utils.emitter import ee
import time
from .prefixer import *
from .connection import Command


class Element:
    tagName: str
    elementId: str
    nodeId: typing.Optional[str]


class Rect(dict):
    key_map = {
        "x": "left",
        "y": "top",
        "w": "width",
        "h": "height",
    }

    def __getattr__(self, name):
        name = Rect.key_map.get(name, name)
        return self[name]

    def __getitem__(self, name):
        name = Rect.key_map.get(name, name)
        return super(Rect, self).__getitem__(name)


class BaseElement(MiniumObject, metaclass=RetryableApi):
    """
    元素基类
    """
    __RETRY_EXCEPTION__ = (MiniTimeoutCauseByConnectionBreakError,)  # 小程序掉线Element实例可能就没用了
    __RETRY_API__ = [
        "data",
        # "size",
        "offset",
        # "rect",
        # "page_scroll_x",
        # "page_scroll_y",
        # "page_offset",
        # "value",
        # "inner_text",
        "inner_wxml",
        "outer_text",
        "styles",
        "attribute",
        "scroll_to",
        # private
        "_property",  # 属性相关
        "_dom_property",  # dom属性相关
        "_get_window_properties",  # window属性相关
        "__get_elements",  # get element相关都经过该接口
    ]

    def __init__(self, element: Element, page_id: str, connection):
        """
        初始化
        """
        super().__init__()
        self.element_id = element.elementId
        self.page_id = page_id
        self._tag_name = element.tagName
        self.connection = connection

    def __str__(self) -> str:
        return "%(class)s[%(eid)s][%(tag)s]" % {
            "class": self.__class__.__name__,
            "eid": self.element_id,
            "tag": self._tag_name,
        }

    def __getattribute__(self, item):
        # element 的魔法糖，获取标签上{name}属性
        try:
            return super(BaseElement, self).__getattribute__(item)
        except AttributeError:
            if not item.startswith("__"):
                prop = self._property(item)[0]
                if prop is not None:
                    return prop
                attr = self.attribute(item)[0]
                if attr is not None:
                    return attr
            raise

    @property
    def data(self):
        """
        自定义组件子类来实现
        :param data:
        :return:
        """
        return

    @data.setter
    def data(self, data):
        """
        自定义组件子类来实现
        :param data:
        :return:
        """
        pass

    @property
    def size(self):
        """
        get element size
        :return: {"width", "height"}
        """
        size_arr = self._dom_property(["offsetWidth", "offsetHeight"])
        return Rect({"width": size_arr[0], "height": size_arr[1]})

    @property
    def offset(self):
        """
        get element offset,
        :return: {"left","top"}
        """
        rtn = self._send("Element.getOffset")
        return Rect(rtn.result)

    @property
    def rect(self):
        """
        get element Rect
        :return: {"left","top","width", "height"}
        """
        rect = self.offset
        rect.update(self.size)
        return rect

    @property
    def page_scroll_x(self):
        """
        获取窗口顶点与页面顶点的 x 轴偏移量
        :return:
        """
        return self._get_window_properties(["scrollX"])[0]

    @property
    def page_scroll_y(self):
        """
        获取窗口顶点与页面顶点的 y 轴偏移量
        :return:
        """
        return self._get_window_properties(["scrollY"])[0]

    @property
    def page_offset(self):
        """
        获取页面的偏移坐标
        """
        ret = self._get_window_properties(["scrollX", "scrollY"])
        return Rect({"left": ret[0], "top": ret[1]})

    @property
    def value(self):
        """
        get element vaule
        :return:
        """
        return self._property("value")[0]

    @property
    def inner_text(self):
        """
        get element text
        :return:
        """
        return self._dom_property("innerText")[0]

    text = inner_text

    @property
    def inner_wxml(self):
        """
        get wxml for element
        :return:
        """
        return self._send("Element.getWXML", {"type": "inner"}).result.wxml

    @property
    def outer_wxml(self):
        """
        get wxml for element self
        :return:
        """
        return self._send("Element.getWXML", {"type": "outer"}).result.wxml

    def styles(self, names):
        """
        get element styles
        :param names:
        :return:
        """
        return self._getter("getStyles", "styles", names)

    def call_method(self, method: str, *params):
        """
        子类来实现
        """
        pass

    def call_func(self, func: str, args=None):
        """
        在 WebView 中执行 JS 脚本，同时会传入 dom 元素
        :param func: 方法名
        :param args: 参数
        :return:
        """
        if not args:
            args = []
        return self._send(
            "Element.callFunction", {"functionName": func, "args": args}
        ).result

    def tap(self, *args, **kwargs):
        """
        点击
        :return: NULL
        """
        self._send("Element.tap")
        ee.emit("notify")

    def click(self, *args, **kwargs):
        """
        点击, 同 tap
        :return: NULL
        """
        styles = self.styles("pointer-events")
        if styles and styles[0] == "none":
            self.logger.warning("can't click, because pointer-events is none")
            return
        self.tap()
        time.sleep(1)

    def move(self, x_offset, y_offset, move_delay=350, smooth=False):
        """
        移动 element
        :param x_offset: x 方向上的偏移，往右为正数，往左为负数
        :param y_offset: y 方向上的偏移，往下为正数，往上为负数
        :param move_delay: 移动延时 (ms)
        :param smooth: 是否平滑移动
        :return:
        """
        self._move(
            x_offset=x_offset, y_offset=y_offset, move_delay=move_delay, smooth=smooth
        )

    def long_press(self, duration=350):
        """
        长按
        :param duration: 时长 (ms)
        :return: NULL
        """
        offset = self.offset
        size = self.size
        page_offset = self.page_offset
        ori_changed_touch = ori_touch = {
            "identifier": 0,
            "pageX": offset["left"] + size["width"] // 2,
            "pageY": offset["top"] + size["height"] // 2,
            "clientX": offset["left"] + size["width"] // 2 - page_offset.x,
            "clientY": offset["top"] + size["height"] // 2 - page_offset.y,
        }
        self._touch_start(touches=[ori_touch], changed_touches=[ori_changed_touch])
        time.sleep(duration / 1000)
        self._touch_end(changed_touches=[ori_changed_touch])
        ee.emit("notify")

    def touch_start(self, touches: list, changed_touches: list):
        """
        touch start
        :param touches:
        :param changed_touches:
        :return:
        """
        self._touch_start(touches=touches, changed_touches=changed_touches)

    def touch_end(self, changed_touches: list):
        """
        touch end
        :param changed_touches:
        :return:
        """
        self._touch_end(changed_touches=changed_touches)

    def touch_move(self, touches: list, changed_touches: list):
        """
        touch move
        :param touches:
        :param changed_touches:
        :return:
        """
        self._touch_move(touches=touches, changed_touches=changed_touches)

    def touch_cancel(self):
        """
        touch cancel
        :return: NULL
        """
        self._send("Element.touchcancel")

    def slide(self, direction, distance):
        """
        拖动
        :param direction:方向
        :param distance:距离
        :return:NULL
        """
        raise NotImplementedError()

    def get_element(
        self, selector, inner_text=None, value=None, text_contains=None, max_timeout=0
    ):
        """
        find elements in current page
        目前支持的选择器有：

        选择器	              样例	                      样例描述
        .class	             .intro	                    选择所有拥有 class="intro" 的组件
        #id	                 #firstname	                选择拥有 id="firstname" 的组件
        tagname	             view	                    选择所有    view 组件
        tagname, tagname	 view, checkbox	            选择所有文档的 view 组件和所有的 checkbox 组件
        ::after	             view::after	            在 view 组件后边插入内容
        ::before	         view::before	            在 view 组件前边插入内容

        :param selector: 选择器
        :param inner_text: inner_text
        :param value: value
        :param text_contains: 包含的文字
        :param max_timeout: 超时时间
        :return:element 对象
        """
        r = self.get_elements(
            selector,
            max_timeout,
            inner_text=inner_text,
            value=value,
            text_contains=text_contains,
            index=0,
        )
        if not r:
            raise MiniElementNotFoundError("element[%s] not found" % selector)
        return r[0]

    def get_elements(
        self,
        selector,
        max_timeout=0,
        inner_text=None,
        text_contains=None,
        value=None,
        index=-1,
    ) -> typing.List[ElementType]:  # type: ignore # noqa: F811
        """
        find elements in current page
        目前支持的选择器有：

        选择器	              样例	                      样例描述
        .class	             .intro	                    选择所有拥有 class="intro" 的组件
        #id	                 #firstname	                选择拥有 id="firstname" 的组件
        tagname	             view	                    选择所有    view 组件
        tagname, tagname	 view, checkbox	            选择所有文档的 view 组件和所有的 checkbox 组件
        ::after	             view::after	            在 view 组件后边插入内容
        ::before	         view::before	            在 view 组件前边插入内容

        :param selector: 选择器
        :param max_timeout: 超时时间
        :return:element 对象 list
        """
        if inner_text is None and value is None and text_contains is None:
            need_filter = False
        else:
            need_filter = True
        if not need_filter:
            return self._get_elements_by_css(selector, max_timeout, index=index)

        @timeout(max_timeout)
        def filter_elements():
            # 需要过滤内容，有返回元素不是终结条件, 需要获取所有元素后再作过滤
            elements = self._get_elements_by_css(selector, max_timeout, index=-1)
            els = []
            for element in elements:
                if inner_text and element.inner_text != inner_text:
                    continue
                if value and element.value() != value:
                    continue
                if text_contains and text_contains not in element.inner_text:
                    continue
                els.append(element)
                if len(els) == (index + 1):
                    return els
            return els

        return filter_elements()

    def __get_elements(self, selector: str, index=-1):
        elements = []
        ret = self._send("Element.getElements", {"selector": selector})
        for el in ret.result.elements:
            elements.append(create(el, self.page_id, self.connection))
            if len(elements) == (index + 1):  # index==-1时，不会match，就会全部返回
                return elements
        return elements

    def __search_child(
        self, selector_list: list, parent: ElementType = None, index=-1  # type: ignore # noqa: F811
    ) -> typing.List[ElementType]:
        # index == -1: get所有, index >=0: get index+1个
        if len(selector_list) == 0:
            return []
        _selector = selector_list.pop()
        should_be_custom_element = bool(
            len(selector_list)
        )  # 出了最后一层selector，都需要是自定义组件，不然不能往下走
        if parent:
            els = parent.get_elements(
                _selector, max_timeout=0
            )  # 调用element.get_elements方法，要求立刻返回，不然会影响效率
        else:
            els = self.__get_elements(_selector)
        if len(els) == 0:
            return []
        real_els = []
        for _el in els:
            if should_be_custom_element and isinstance(_el, CustomElement):
                real_els.append(_el)
            elif not should_be_custom_element:
                real_els.append(_el)
            else:
                self.logger.warn("%s should be a custom element" % _selector)
        if not real_els or len(selector_list) == 0:  # 找不到 或 没有子选择器了
            return real_els
        child_els = []
        for _el in real_els:
            child_el = self.__search_child(
                selector_list[0:], _el
            )  # selector_list[0:]相当于copy
            child_els += child_el
            if len(child_els) == (index + 1):
                return child_els
        return child_els

    def _get_elements_by_css(self, selector: str, max_timeout=0, index=-1) -> typing.List[ElementType]:
        """
        1. 现存自定义组件中的class会自动加前缀，但不包括slot占位的元素
        2. slot元素的class命名规则
        2.1 <page><test><view class="这个class不会加前缀"></view></test></page>
        2.2 <custom><test><view class="这个class会加上custom组件对应的前缀"></view></test></custom>
        3. 自定义组件中通过id获取元素失败
        """

        @timeout(max_timeout)
        def search_elements(_selector: list or tuple):
            return self.__search_child(_selector[0:], index=index)

        self.logger.info("try to get elements: %s" % selector)
        _selector_list = selector.split(">>>")
        _selector_list.reverse()  # 新增处理【>>>】穿透自定义组件逻辑
        els = search_elements(_selector_list)
        if len(els) == 0:
            self.logger.warning(f"Could not found any element '{selector}' you need")
        else:
            self.logger.info("find elements success: %s" % str(els))
        return els

    def attribute(self, name):
        """
        获取元素属性
        :return: attribute
        """
        return self._getter("getAttributes", "attributes", name)

    def trigger(self, trigger_type, detail):
        """
        just a trigger
        :param trigger_type:
        :param detail:
        :return:
        """
        return self._trigger(trigger_type, detail)

    def trigger_events(self, events):
        """触发DOM事件序列, 录制回放用
        touch事件可以只提供type, interval参数

        :param [{type, interval?, touches?, changedTouches?, detail?}...] events: 事件列表
        """
        for event in events:
            interval = event.get("interval", 0)
            if event["type"] == "touchstart":
                self._touch_start(event["touches"], event["changedTouches"])
            elif event["type"] == "touchmove":
                self._touch_move(event["touches"], event["changedTouches"])
            elif event["type"] == "touchend":
                self._touch_end(event["changedTouches"])
            else:
                self.trigger(event["type"], event["detail"])
            time.sleep(interval)

    def dispatch_event(
        self,
        event_name,
        selector=None,
        touches=None,
        change_touches=None,
        detail=None,
        **kwargs,
    ):
        params = {"eventName": event_name, "eventData": {}}
        if selector:
            params["$"] = selector
        if touches:
            params["eventData"]["touches"] = touches
        if change_touches:
            params["eventData"]["changeTouches"] = change_touches
        if detail:
            params["eventData"]["detail"] = detail
        if kwargs:
            params["eventData"].update(kwargs)
        return self._send("Element.dispatchEvent", params)

    def scroll_to(self, top, left):
        return self.dispatch_event(
            "scroll", scrollDetail={"scrollTop": top, "scrollLeft": left}
        )

    # private method

    def _property(self, name):
        """
        get property
        :param name:
        :return:
        """
        return self._getter("getProperties", "properties", name)

    def _dom_property(self, name):
        """
        get property from dom
        :param name:
        :return:
        """
        return self._getter("getDOMProperties", "properties", name)

    def _get_window_properties(self, names=None):
        """
        获取 window 对象的属性值。
        :param names:
        :return:
        """
        if names is None:
            names = []
        return self._send(
            "Page.getWindowProperties", {"names": names}
        ).result.properties

    def _trigger(self, trigger_type, detail):
        params = dict()
        params["type"] = trigger_type
        if detail:
            params["detail"] = detail
        return self._send("Element.triggerEvent", params)

    def _getter(self, method, return_name, names=""):
        if isinstance(names, list):
            result = self._send("Element." + method, {"names": names})
        elif isinstance(names, str):
            result = self._send("Element." + method, {"names": [names]})
        else:
            raise Exception("invalid names type")
        ret = getattr(result.result, return_name)
        return ret

    def _send(self, method, params=None, extra_time=0):
        if params is None:
            params = {}
        params["elementId"] = self.element_id
        params["pageId"] = self.page_id

        return self.connection.send(method, params, max_timeout=Command.max_timeout+extra_time)

    def _move(self, x_offset, y_offset, move_delay=350, smooth=False):
        offset = self.offset
        size = self.size
        page_offset = self.page_offset
        changed_touch = touch = ori_changed_touch = ori_touch = {
            "identifier": 0,
            "pageX": offset["left"] + size["width"] // 2,
            "pageY": offset["top"] + size["height"] // 2,
            "clientX": offset["left"] + size["width"] // 2 - page_offset.x,
            "clientY": offset["top"] + size["height"] // 2 - page_offset.y,
        }
        self._touch_start(touches=[ori_touch], changed_touches=[ori_changed_touch])
        # time.sleep(move_delay / 1000)
        if smooth and (x_offset or y_offset):  # offset不能都为0
            time.sleep(move_delay / 4000)
            temp_x_offset = temp_y_offset = 0
            max_offset = max(abs(x_offset), abs(y_offset))
            step = (move_delay / 2000) / max_offset
            while abs(temp_x_offset) <= abs(x_offset) or abs(temp_y_offset) <= abs(
                y_offset
            ):
                if temp_x_offset == x_offset:
                    pass
                elif not x_offset == 0:
                    temp_x_offset = (
                        (temp_x_offset + 1) if x_offset > 0 else (temp_x_offset - 1)
                    )
                if temp_y_offset == y_offset:
                    pass
                elif not y_offset == 0:
                    temp_y_offset = (
                        (temp_y_offset + 1) if y_offset > 0 else (temp_y_offset - 1)
                    )

                changed_touch = touch = {
                    "identifier": 0,
                    "pageX": offset["left"] + size["width"] // 2 + temp_x_offset,
                    "pageY": offset["top"] + size["height"] // 2 + temp_y_offset,
                    "clientX": offset["left"]
                    + size["width"] // 2
                    - page_offset.x
                    + temp_x_offset,
                    "clientY": offset["top"]
                    + size["height"] // 2
                    - page_offset.y
                    + temp_y_offset,
                }
                self._touch_move(touches=[touch], changed_touches=[changed_touch])
                if temp_x_offset == x_offset and temp_y_offset == y_offset:
                    break
                time.sleep(step)
            time.sleep(move_delay / 4000)
        else:
            time.sleep(move_delay / 2000)
            changed_touch = touch = {
                "identifier": 0,
                "pageX": offset["left"] + size["width"] // 2 + x_offset,
                "pageY": offset["top"] + size["height"] // 2 + y_offset,
                "clientX": offset["left"]
                + size["width"] // 2
                - page_offset.x
                + x_offset,
                "clientY": offset["top"]
                + size["height"] // 2
                - page_offset.y
                + y_offset,
            }
            self._touch_move(touches=[touch], changed_touches=[changed_touch])
            time.sleep(move_delay / 2000)
        # time.sleep(move_delay / 1000)
        self._touch_end(changed_touches=[changed_touch])

    def _touch_start(self, touches: list, changed_touches: list):
        self._send(
            "Element.touchstart",
            params={"touches": touches, "changedTouches": changed_touches},
        )

    def _touch_move(self, touches: list, changed_touches: list):
        self._send(
            "Element.touchmove",
            params={"touches": touches, "changedTouches": changed_touches},
        )

    def _touch_end(self, changed_touches: list):
        touches = []
        self._send(
            "Element.touchend",
            params={"touches": touches, "changedTouches": changed_touches},
        )

    # def _get_elements(self, selector, max_timeout=0, index=-1):
    #     @timeout(max_timeout)
    #     def search_elements(_selector):
    #         elements = []
    #         ret = self._send("Element.getElements", {"selector": _selector})
    #         if hasattr(ret, "error"):
    #             raise Exception(
    #                 "Error when finding elements[%s], %s" % (_selector, ret.error)
    #             )
    #         for el in ret.result.elements:
    #             elements.append(create(el, self.page_id, self.connection))
    #         return elements

    #     def search_child(
    #         selector_list: list, parent: BaseElement = None
    #     ):  # index == -1: get所有, index >=0: get index+1个
    #         if len(selector_list) == 0:
    #             return
    #         _selector = selector_list.pop()
    #         should_be_custom_element = bool(
    #             len(selector_list)
    #         )  # 出了最后一层selector，都需要是自定义组件，不然不能往下走
    #         if parent:
    #             els = parent.get_elements(_selector)
    #         else:
    #             els = search_elements(_selector)
    #         if len(els) == 0:
    #             return []
    #         real_els = []
    #         for _el in els:
    #             if should_be_custom_element and isinstance(_el, CustomElement):
    #                 real_els.append(_el)
    #             elif not should_be_custom_element:
    #                 real_els.append(_el)
    #             else:
    #                 self.logger.warn("%s should be a custom element" % _selector)
    #         if not real_els or len(selector_list) == 0:  # 找不到 或 没有子选择器了
    #             return real_els
    #         child_els = []
    #         for _el in real_els:
    #             child_el = search_child(selector_list[0:], _el)
    #             child_els += child_el
    #             if index == -1:
    #                 continue
    #             elif len(child_els) > index:
    #                 return child_els
    #         return child_els

    #     try:
    #         self.logger.info("try to find elements: %s" % selector)
    #         _selector_list = selector.split(">>>")
    #         if len(_selector_list) > 1:  # 新增处理【>>>】穿透自定义组件逻辑
    #             _selector_list.reverse()
    #             els = search_child(_selector_list)
    #         else:
    #             els = search_elements(selector)
    #         if len(els) == 0:
    #             self.logger.warning(
    #                 f"Could not found any element '{selector}' you need"
    #             )
    #         else:
    #             self.logger.info("find elements success: %s" % str(els))
    #         return els
    #     except Exception as e:
    #         raise Exception("elements search fail cause: " + str(e))


class FormElement(BaseElement):
    """
    表单类型元素
    """

    def __init__(self, element, page_id, connection):
        """
        初始化
        """
        super(FormElement, self).__init__(element, page_id, connection)

    #####################
    #       input       #
    #####################
    def input(self, text: str):
        """
        input 标签输入文本
        :param text:
        :return:
        """
        if self._tag_name != "input" and self._tag_name != "textarea":
            self.logger.warning(
                "Element's type is not fit for the method which you call"
            )
            return
        func = "{x}.input".format(x=self._tag_name)
        value = text.strip()
        self.call_func(func, args=[value])
        if text.endswith("\n"):  # 换行符结尾认为是confirm
            self.trigger("confirm", {"value": value})
        # self.trigger("input", {"value": text})

    #####################
    #       picker      #
    #####################
    def pick(self, value):
        """
        处理 picker 组件
        picker 的类型:{
            selector: 普通选择器 => value(int) 表示选择了 range 中的第几个 (下标从 0 开始) ,
            multiSelector: 多列选择器 => value(int) 表示选择了 range 中的第几个 (下标从 0 开始)  ,
            time: 时间选择器 => value(str) 表示选中的时间，格式为"hh:mm"
            date: 日期选择器 => value(str) 表示选中的日期，格式为"YYYY-MM-DD"
            region: 省市区选择器 => value(int) 表示选中的省市区，默认选中每一列的第一个值
            }
        :param value: 需要选择的选项
        :return:
        """
        if self._tag_name != "picker":
            self.logger.warning(
                "Element's type is not fit for the method which you call"
            )
            return
        self.trigger("change", {"value": value})
        return
    
    def tap(self, force=False, **kwargs):
        """input/picker点击事件会触发拉起键盘/选择器

        :param bool force: 组件绑定了bindtap事件, 一定需要触发tap操作, defaults to False
        :return None: 
        """
        if self._tag_name in ("picker", "input") and not force:
            return
        return super().tap(**kwargs)

    def click(self, force=False):
        if self._tag_name in ("picker", "input") and not force:
            return
        return super(FormElement, self).click()

    #####################
    #       switch      #
    #####################
    def switch(self):
        """
        点击改变 switch 的状态
        :return:
        """
        if self._tag_name != "switch":
            self.logger.warning(
                "Element's type is not fit for the method which you call"
            )
            return
        func = "switch.tap"
        self.call_func(func, args=[])

    #####################
    #       slider      #
    #####################
    def slide_to(self, value):
        """
        slider 组件滑动到指定数值
        :return:
        """
        if self._tag_name != "slider":
            self.logger.warning(
                "Element's type is not fit for the method which you call"
            )
            return
        func = "slider.slideTo"
        self.call_func(func, args=[value])


class ViewElement(BaseElement):
    """
    视图类型元素
    """

    def __init__(self, element, page_id, connection):
        """
        初始化
        """
        super(ViewElement, self).__init__(element, page_id, connection)

    #####################
    #   scroll-view     #
    #####################
    def scroll_to(self, x=0, y=0):
        """
        scroll-view 滚动指定距离
        :param x: x 轴上的距离
        :param y: y 轴上的距离
        :return:
        """
        if self._tag_name != "scroll-view":
            self.logger.warning(
                "Element's type is not fit for the method which you call"
            )
            return
        func = "scroll-view.scrollTo"
        self.call_func(func, args=[x, y])

    @property
    def scroll_left(self):
        """
        scroll-view scrollLeft
        :return:
        """
        if self._tag_name != "scroll-view":
            self.logger.warning(
                "Element's type is not fit for the method which you call"
            )
            return
        func = "scroll-view.scrollLeft"
        return self.call_func(func)

    @property
    def scroll_top(self):
        """
        scroll-view scrollTop
        :return:
        """
        if self._tag_name != "scroll-view":
            self.logger.warning(
                "Element's type is not fit for the method which you call"
            )
            return
        func = "scroll-view.scrollTop"
        return self.call_func(func)

    @property
    def scroll_width(self):
        """
        scroll-view scrollWidth
        :return:
        """
        if self._tag_name != "scroll-view":
            self.logger.warning(
                "Element's type is not fit for the method which you call"
            )
            return
        func = "scroll-view.scrollWidth"
        self.call_func(func)

    @property
    def scroll_height(self):
        """
        scroll-view 滚动指定距离
        :return:
        """
        if self._tag_name != "scroll-view":
            self.logger.warning(
                "Element's type is not fit for the method which you call"
            )
            return
        func = "scroll-view.scrollHeight"
        self.call_func(func)

    #####################
    #       swiper      #
    #####################
    def swipe_to(self, index):
        """
        切换滑块视图容器当前的页面
        :param index:
        :return:
        """
        if self._tag_name != "swiper":
            self.logger.warning(
                "Element's type is not fit for the method which you call"
            )
            return
        func = "swiper.swipeTo"
        self.call_func(func, args=[index])
        self._trigger("animationfinish", {"current": index, "currentItemId": "", "source": "touch"})

    def trigger(self, trigger_type, detail):
        if trigger_type == "animationfinish":  # swiper会自动触发
            return
        return super().trigger(trigger_type, detail)

    ########################
    #       movable-view   #
    ########################
    def move_to(self, x, y):
        """
        可移动的视图容器拖拽滑动
        :param x: x轴方向的偏移
        :param y: y轴方向的偏移
        :return:
        """
        if self._tag_name != "movable-view":
            self.logger.warning(
                "Element's type is not fit for the method which you call"
            )
            return
        func = "movable-view.moveTo"
        self.call_func(func, args=[x, y])


class VideoElement(BaseElement):
    """
    视频播放类型元素
    """

    def __init__(self, element, page_id, connection):
        """
        初始化
        """
        super(VideoElement, self).__init__(element, page_id, connection)
        self.video_id = element.get("videoId", None)
        self.media_type = MediaType.VIDEO

    def _call_context_method(self, method: str, args: list = None):
        if not args:
            args = []
        params = {"videoId": self.video_id, "method": method, "args": args}
        return self._send("Element.callContextMethod", params=params)

    def play(self):
        """
        播放视频
        :return:
        """
        # self.controller.play()
        self._call_context_method(method="play")

    def pause(self):
        """
        暂停视频
        :return:
        """
        # self.controller.pause()
        self._call_context_method(method="pause")

    def stop(self):
        """
        停止视频
        :return:
        """
        # self.controller.stop()
        self._call_context_method(method="stop")

    def seek(self, position: int):
        """
        跳转到指定位置
        :param position: 跳转到的位置，单位 s
        :return:
        """
        # self.controller.seek(position)
        self._call_context_method(method="seek", args=[position])

    def send_danmu(self, text: str, color="#ff0000"):
        """
        发送弹幕
        :param text: 弹幕文字
        :param color: 弹幕颜色
        :return:
        """
        # self.controller.send_danmu(text, color)
        self._call_context_method(
            method="sendDanmu", args=[{"text": text, "color": color}]
        )

    def playback_rate(self, rate: float):
        """
        设置倍速播放
        :param rate: 倍率，支持 0.5/0.8/1.0/1.25/1.5，2.6.3 起支持 2.0 倍速
        :return:
        """
        # self.controller.playback_rate(rate)
        self._call_context_method(method="playbackRate", args=[rate])

    def request_full_screen(self, direction=0):
        """
        全屏
        :param direction: 设置全屏时视频的方向，不指定则根据宽高比自动判断
        0	正常竖向
        90	屏幕逆时针90度
        -90	屏幕顺时针90度
        :return: status
        """
        # self.controller.request_full_screen(direction)
        self._call_context_method(
            method="requestFullScreen", args=[{"direction": direction}]
        )

    def exit_full_screen(self):
        """
        退出全屏
        :return:
        """
        # self.controller.exit_full_screen()
        self._call_context_method(method="exitFullScreen")

    def show_status_bar(self):
        """
        显示状态栏，仅在iOS全屏下有效
        :return:
        """
        # self.controller.show_status_bar()
        self._call_context_method(method="showStatusBar")

    def hide_status_bar(self):
        """
        隐藏状态栏，仅在iOS全屏下有效
        :return:
        """
        # self.controller.hide_status_bar()
        self._call_context_method(method="hideStatusBar")


class AudioElement(BaseElement):
    """
    音频播放类型元素
    """

    def __init__(self, element, page_id, connection):
        """
        初始化
        """
        super(AudioElement, self).__init__(element, page_id, connection)
        self.controller = AudioController(self)
        self.media_type = MediaType.AUDIO

    def set_src(self, src):
        """
        设置音频地址
        :param src: 音频地址
        :return:
        """
        self.controller.setSrc(src)

    def play(self):
        """
        播放音频
        :return:
        """
        self.controller.play()

    def pause(self):
        """
        暂停音频
        :return:
        """
        self.controller.pause()

    def seek(self, position):
        """
        跳转到指定位置
        :param position: 时间，单位：s
        :return:
        """
        self.controller.seek(position)


class LivePlayerElement(BaseElement):
    """
    直播播放类型元素
    """

    def __init__(self, element, page_id, connection):
        """
        初始化
        """
        super(LivePlayerElement, self).__init__(element, page_id, connection)
        self.controller = LivePlayerController(self)
        self.media_type = MediaType.LIVE_PLAY

    def play(self):
        """
        播放
        :return:
        """
        self.controller.play()

    def stop(self):
        """
        停止
        :return:
        """
        self.controller.stop()

    def mute(self):
        """
        静音
        :return:
        """
        self.controller.mute()

    def pause(self):
        """
        暂停
        :return:
        """
        self.controller.pause()

    def resume(self):
        """
        恢复
        :return:
        """
        self.controller.resume()

    def request_full_screen(self, direction=0):
        """
        全屏
        :param direction: 方向
        0	正常竖向
        90	屏幕逆时针90度
        -90	屏幕顺时针90度
        :return: status
        """
        self.controller.requestFullScreen(direction)

    def exit_full_screen(self):
        """
        退出全屏
        :return:
        """
        self.controller.exitFullScreen()

    def snapshot(self):
        """
        截屏
        :return:
        """
        self.controller.snapshot()

    def request_picture_in_picture(self):
        """
        画中画，2.15.0后支持
        :return:
        """
        self.controller.requestPictureInPicture()

    def exit_picture_in_picture(self):
        """
        退出画中画
        :return:
        """
        self.controller.exitPictureInPicture()


class LivePusherElement(BaseElement):
    """
    直播推流类型元素
    """

    def __init__(self, element, page_id, connection):
        """
        初始化
        """
        super(LivePusherElement, self).__init__(element, page_id, connection)
        self.controller = LivePusherController(self)
        self.media_type = MediaType.LIVE_PUSH

    def start(self):
        """
        开始推流，同时开启摄像头预览
        :return:
        """
        self.controller.start()

    def stop(self):
        """
        停止推流，同时停止摄像头预览
        :return:
        """
        self.controller.stop()

    def pause(self):
        """
        暂停推流
        :return:
        """
        self.controller.pause()

    def resume(self):
        """
        恢复推流
        :return:
        """
        self.controller.resume()

    def switch_camera(self):
        """
        切换前后摄像头
        :return:
        """
        self.controller.switchCamera()

    def snapshot(self):
        """
        截屏
        :return:
        """
        self.controller.snapshot()

    def toggle_torch(self):
        """
        切换手电筒
        :return:
        """
        self.controller.toggleTorch()

    def play_bgm(self, url):
        """
        播放背景音
        :param url: 背景音链接
        :return:
        """
        self.controller.playBGM({"url": url})

    def stop_bgm(self):
        """
        停止背景音
        :return:
        """
        self.controller.stopBGM()

    def pause_bgm(self):
        """
        暂停背景音
        :return:
        """
        self.controller.pauseBGM()

    def resume_bgm(self):
        """
        恢复背景音
        :return:
        """
        self.controller.resumeBGM()

    def set_bgm_volume(self, volume):
        """
        设置背景音音量
        :param volume: 音量大小，范围是 0-1
        :return:
        """
        self.controller.setBGMVolume({"volume": volume})

    def set_mic_volume(self, volume):
        """
        设置麦克风音量
        :param volume: 音量大小，范围是 0-1
        :return:
        """
        self.controller.setMICVolume({"volume": volume})

    def start_preview(self):
        """
        开启摄像头预览
        :return:
        """
        self.controller.startPreview()

    def stop_preview(self):
        """
        关闭摄像头预览
        :return:
        """
        self.controller.stopPreview()

    def send_message(self, msg):
        self.controller.sendMessage({"msg": msg})


class CustomElement(BaseElement):
    def __init__(self, element, page_id, connection):
        """
        初始化
        """
        super(CustomElement, self).__init__(element, page_id, connection)
        self.node_id = element.get("nodeId", None)

    def _send(self, method, params=None):
        if params is None:
            params = {}
        params["elementId"] = self.element_id
        params["pageId"] = self.page_id
        params["nodeId"] = self.node_id
        return self.connection.send(method, params)

    @property
    def data(self):
        return self._send("Element.getData").result.data

    @data.setter
    def data(self, data):
        """
        设置页面 data
        :param data:
        :return:
        """
        self._send("Element.setData", {"data": data})

    def call_method(self, method: str, *params):
        """
        调用自定义组件实例方法
        :param method:
        :param params:
        :return:
        """
        if not params:
            params = []
        return self._send("Element.callMethod", {"method": method, "args": params})


class MediaController(object):
    """
    媒体类型元素控制
    1. 使用 `wx.createXXXContext` 接口创建 context （还有一种是通过 wx.createSelectorQuery().select("#ID").context获取）
    2. 以"{pageId}_{elementId}"为key记录context
    3. hook page.onUnload, 页面销毁时需要清除
    4. 通过evaluate调用context的方法
    TODO: plugin中不支持evaluate
    """

    create_context_method = ""  # `wx.createXXXContext` 中的 `createXXXContext`
    allow_method = ()  # controller支持的方法与context上的一致

    def __init__(self, element: BaseElement):
        """
        初始化
        """
        self.element = element
        self.element_id = element.attribute("id")[0]
        # if not self.create_context_method:
        #     raise Exception("%s.create_context_method will not be empty" % self.__class__.__name__)
        self.create_context()

    def create_context(self):
        # 以"{pageId}_{elementId}"为key创建context, elementId是标签上的ID
        # hook page.onUnload, 页面销毁时需要清除
        self.element._evaluate_js(
            "createContext",
            [self.element_id, self.element.page_id],
            code_format_info={"createMediaContext": self.create_context_method},
        )

    def _call_context_method(self, method: str, args: list = None):
        if not args:
            args = []
        if not isinstance(args, list):
            args = [args]
        return self.element._evaluate_js(
            "callContextMethod",
            [self.element_id, self.element.page_id, method, args],
            code_format_info={"createMediaContext": self.create_context_method},
            desc="evaluate %s.%s" % (self.create_context_method[6:], method),
        )

    # 透传context的方法
    def __getattribute__(self, name: str) -> typing.Any:
        try:
            return super().__getattribute__(name)
        except AttributeError:
            if name in self.allow_method:
                return lambda *args, **kwargs: self._call_context_method(
                    name, *args, **kwargs
                )
            raise


class AudioController(MediaController):
    """
    音频播放控制
    """

    create_context_method = "createAudioContext"
    allow_method = (
        "setSrc",
        "play",
        "pause",
        "seek",
    )


class LivePlayerController(MediaController):
    """
    直播播放控制
    """

    create_context_method = (
        "createLivePlayerContext"  # `wx.createXXXContext` 中的 `createXXXContext`
    )
    allow_method = (
        "play",
        "stop",
        "mute",
        "pause",
        "requestFullScreen",
        "exitFullScreen",
        "requestPictureInPicture",
        "exitPictureInPicture",
        "resume",
        "snapshot",
    )


class LivePusherController(MediaController):
    """
    直播播放/推流控制
    """

    create_context_method = (
        "createLivePusherContext"  # `wx.createXXXContext` 中的 `createXXXContext`
    )
    allow_method = (
        "pause",
        "pauseBGM",
        "playBGM",
        "resume",
        "resumeBGM",
        "sendMessage",
        "setBGMVolume",
        "setMICVolume",
        "snapshot",
        "start",
        "startPreview",
        "stop",
        "stopBGM",
        "stopPreview",
        "switchCamera",
        "toggleTorch",
    )


ELEMENT_TYPE = {
    "video": VideoElement,
    "audio": AudioElement,
    "live-player": LivePlayerElement,
    "live-pusher": LivePusherElement,
    "scroll-view": ViewElement,
    "swiper": ViewElement,
    "movable-view": ViewElement,
    "input": FormElement,
    "textarea": FormElement,
    "switch": FormElement,
    "slider": FormElement,
    "picker": FormElement,
}


ElementType = typing.Union[
    BaseElement,
    CustomElement,
    VideoElement,
    AudioElement,
    ViewElement,
    FormElement,
    LivePlayerElement,
    LivePusherElement,
]


def create(element: Element, page_id: str, connection) -> ElementType:
    if "nodeId" in element.keys():
        return CustomElement(element, page_id, connection)
    else:
        return ELEMENT_TYPE.get(element.tagName, BaseElement)(
            element, page_id, connection
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Created by xiazeng on 2018/5/8
import logging

logger = logging.getLogger()

ELEMENT_OPERATION = "elementOperation"
SCROLL_TO = "scrollTo"
DUMP_DOM_TO_JSON = "dumpDomJson"
CLICK_ELEMENT = "clickElement"


class JsApiError(RuntimeError):
    pass


class CallJsFailed(JsApiError):
    pass


class ElementNotExists(JsApiError):
    pass


class CallJsError(JsApiError):
    pass


class JsResponse(object):
    class ErrorCode:
        success = 0
        unknown = -1
        failed = 1
        error = 2
        notExists = 3

    def __init__(self, json_dict):
        self.ret = json_dict["ret"]
        self.msg = json_dict["msg"]
        self.data = json_dict["data"]

    def is_ok(self):
        return self.ret == JsResponse.ErrorCode.success

    def raise_error(self):
        if not self.is_ok():
            logger.error(self.data)
            if self.ret == JsResponse.ErrorCode.notExists:
                raise ElementNotExists(self.msg)
            elif self.ret == JsResponse.ErrorCode.failed:
                raise CallJsFailed(self.msg)
            elif self.ret == JsResponse.ErrorCode.error:
                raise CallJsError(self.msg)
            else:
                raise CallJsFailed(self.msg)


class ELEMENT_OPERATION_SUPPLY_METHOD:
    SET_PROPERTY = "_setProperty"
    GET_PROPERTY = "_getProperty"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Created by xiazeng on 2017/8/25
import inspect
import json
import logging
import time
import os.path

import at.core.adbwrap as adbwrap
import at.utils.commonhelper as commonhelper
import requests

from .exceptions import *

logger = logging.getLogger()


def raise_from_res(ret, msg, data):
    from future.utils import raise_from
    if ret == 0:
        return
    if "stack" in data:
        logger.error(data["stack"])
    elif "result" in data:
        logger.error(data["result"])
    else:
        logger.error(json.dumps(data, indent=3))

    if ret == 3:
        raise_from(NoSuchMethodError(msg), None)
    elif ret == 5:
        raise_from(UiNotFoundError(msg), None)
    elif ret == 7:
        raise_from(UiAutomatorDisconnectError(msg), None)
    elif ret == 4:
        raise_from(ParamError(msg), None)
    else:
        raise_from(RemoteError(msg), None)


class JavaBaseDriver(object):
    SERVER_PORT = 9999
    instance_cache = {}

    def __init__(self, serial):
        name = self.unique_name(serial)
        if name in JavaBaseDriver.instance_cache:
            raise AtError("%s has init, you should call apply to create a object")
        self.serial = serial
        adb = adbwrap.AdbWrap.apply_adb(serial)
        self._last_error = None
        self.adb = adb
        self._port = commonhelper.pick_unuse_port()
        self._app_server_is_run = False
        self._cmd_process = None
        self._start_ts = time.time()
        self.req_costs = []
        self.max_init_retry = 100
        self.max_request_timeout = 60
        self._reconnect_when_error = True
        self.retry_count = 2

        self.hook_list = []
        self.requests_when_reconnect = []

    def add_request_when_reconnect(self, action, http_params=None, **kwargs):
        """
        添加重启at的时候会发送的请求
        """
        self.requests_when_reconnect.append((action, http_params, kwargs))

    def set_reconnect_when_error(self, true_or_false):
        self._reconnect_when_error = true_or_false

    def reconnect(self):
        pass

    def register(self, hook):
        self.hook_list.append(hook)

    def trigger(self, event_name, *args, **kwargs):
        for hook in self.hook_list:
            if getattr(hook, event_name):
                getattr(hook, event_name)(*args, **kwargs)

    def _init(self):
        pass

    def is_remote_running(self):
        return self._app_server_is_run

    def wait_remote_finish(self, timeout=2):
        t = time.time()
        while time.time() - t < timeout:
            if not self.is_remote_running():
                return True
            time.sleep(0.01)
        logger.error("not finish")
        return False

    def set_app_server_run(self, true_or_false):
        logger.debug("true_or_false:%s", true_or_false)
        self._app_server_is_run = true_or_false

    @classmethod
    def apply(cls, serial):
        raise NotImplementedError()

    @classmethod
    def unique_name(cls, name):
        return "%s(%s)" % (cls.__name__, name)

    @classmethod
    def get_cache(cls, serial):
        name = cls.unique_name(serial)
        if name in cls.instance_cache:
            logger.info("get cache %s = %d", name, id(cls.instance_cache[name]))
            return cls.instance_cache[name]
        return None

    @classmethod
    def cache(cls, obj):
        name = obj.page_name(obj.serial)
        if name not in cls.instance_cache:
            logger.info("set cache %s = %d", name, id(obj))
            cls.instance_cache[name] = obj

    @classmethod
    def destroy(cls, serial):
        name = cls.unique_name(serial)
        logger.info(name)
        if name in cls.instance_cache:
            jd = cls.instance_cache[name]
            del cls.instance_cache[name]
            jd.close()

    @staticmethod
    def __redo_ret(res, costs=None):
        if res:
            if res.status_code == requests.codes.ok:
                ret = json.loads(res.text)
                logger.debug("%s java request costs %sms", res.text[:128], costs)
                assert "ret" in ret
                assert "msg" in ret
                assert "data" in ret
                errno = ret["ret"]
                raise_from_res(errno, ret["msg"], ret["data"])
                return ret["data"]
            else:
                raise AtUnknownError(u"pc与手机通讯异常，异常码：" + str(res.status_code))
        else:
            raise AtUnknownError(u"返回结果为空，理论上不应该出现的！" + str(res.status_code))

    @property
    def remote_url(self):
        return "http://127.0.0.1:%d/" % self._port

    def _check_env(self):
        if not self.is_remote_running():
            if self._last_error:
                raise AtError(self._last_error)
            else:
                if self.max_init_retry > 0:
                    self.max_init_retry -= 1
                    self._init()
                else:
                    raise FailedConnectAtServer("can not connect to AtServer")
        # if not self._cmd_process:
        #     raise FailedConnectAtServer("AtServer process was stopped")

        if not self.adb.is_connected():
            raise AdbDisconnectedError("adb has disconnected")

    def do_request(self, action, http_params=None, method="GET", timeout=None, **kwargs):
        if not timeout:
            timeout = self.max_request_timeout
        t = time.time()
        ret = self.request(method.lower(), action, http_params, timeout=timeout, **kwargs)
        if len(self.req_costs) > 1024:
            self.req_costs.pop()
        self.req_costs.append((action, t, time.time()))
        res = ret["result"] if ret is not None and "result" in ret else None
        return res

    def no_check_request(self, action, method="GET", params=None, **kwargs):
        url = self.remote_url + action
        logger.debug("%s, params:%s", url, params)
        session = requests.Session()
        session.trust_env = False
        res = session.request(method, url, params=params, **kwargs)
        session.close()
        return self.__redo_ret(res)

    def request(self, method, action, params=None, **kwargs):
        self._check_env()
        last_exception = None
        s = time.time()

        frames = inspect.stack()
        call = ""
        if len(frames) > 2:
            current_filename = frames[0][1]
            for frame in frames[1:]:
                ignore_files = ["monitor.py", 'case_repair_adapter.py']
                if frame[1] != current_filename and os.path.basename(frame[1]) not in ignore_files and "at/core" not in \
                        frame[1]:
                    call = ", called by %s in line %d" % (frame[3], frame[2])
                    break
        log_prefix = "path:%s, params:%s %s" % (action, json.dumps(params, ensure_ascii=False), call)

        for i in range(2):
            session = requests.Session()
            session.trust_env = False
            try:
                start_time = time.time()
                try:
                    url = self.remote_url + action
                    logger.debug(log_prefix)
                    res = session.request(method, url, params=params, **kwargs)
                    session.close()
                    costs = "%d" % ((time.time() - s) * 1000,)
                    return self.__redo_ret(res, costs)
                except (UiNotFoundError, UiAutomatorDisconnectError):
                    # 重启uiautomator重试一次
                    self.reconnect()
                    logger.debug(log_prefix)
                    res = session.request(method, url, params=params, **kwargs)
                    costs = "%d" % ((time.time() - s) * 1000,)
                    return self.__redo_ret(res, costs)

            except (UiNotFoundError, UiAutomatorDisconnectError):
                raise
            except (requests.ConnectionError,) as e:
                if action == "ping":
                    return {"result": False}
                check_status = self.check_forward()
                if self._cmd_process:
                    uiautomator_exists = True if self._cmd_process.poll() is None else False
                    logger.warning("%s, %d, forward:%s, uiautomator exists:%s, poll:%s, max_init_retry:%s, error:%s",
                                   action, i, check_status, uiautomator_exists, self._cmd_process.poll(),
                                   self.max_init_retry, e)

                last_exception = e
                if self.max_init_retry > 0:
                    self.max_init_retry -= 1
                    self.reconnect()
            except requests.Timeout as e:
                costs = time.time() - start_time
                logging.error("%s, 请求超时%ds，可能原因：\n\t- 有UI刷新\n\t- 底层逻辑超时未返回\n\t- 测试进程被禁用网络（目前vivo手机有发现）", log_prefix,
                              costs)
                self.reconnect()
                last_exception = e
            except Exception as e:
                logging.error("%s, 未知原因", log_prefix)
                last_exception = e
            time.sleep(1 + i * 2)
        else:
            logger.error("%s, 网络异常，超过重试次数:%d，adb连接状态:%s", log_prefix, self.retry_count, self.adb.is_connected())
            raise last_exception

    def check_forward(self):
        if not self.adb.is_connected():
            return False
        is_forward = False
        if str(self._port) in self.adb.run_adb("forward --list"):
            is_forward = True
        # 用别的端口，有
        self.adb.forward_remove(self._port)
        self._port = commonhelper.pick_unuse_port()
        self.adb.forward(self._port, self.SERVER_PORT)
        logger.info("forward status %s", is_forward)
        return is_forward

    def close(self, timeout=5):
        if not self.is_remote_running():
            return
        self.close_remote()
        s = time.time()
        if not self._cmd_process:
            return
        while time.time() - s < timeout:
            if self._cmd_process.poll() is not None or not self.is_remote_running():
                break
            time.sleep(1)
        else:
            self.adb.kill_pid(self._cmd_process.pid)

    def close_remote(self):
        pass

    def print_req_costs(self):
        a = {}
        for req, start, end in self.req_costs:
            if req not in a:
                a[req] = []
            a[req].append((end-start)*1000)
        for req, costs in a.items():
            logger.info("%s num:%s avg_costs:%dms", req, len(costs), sum(costs)/len(costs))

    def __del__(self):
        JavaBaseDriver.close_remote(self.serial)




#!/usr/local/bin/python3
# -*- coding: utf-8 -*-
"""
Author:         yopofeng
Filename:       injectjs.py
Create time:    2021/9/24 21:21
Description:    获取注入的js文件

"""
import os
import re
from enum import Enum
import copy


JS_PATH = os.path.join(os.path.dirname(__file__), "js")
MIN_JS_PATH = os.path.join(JS_PATH, "min")
ES5_JS_PATH = os.path.join(JS_PATH, "es5")  # 只是es5语法的


# 真机调试2.0只支持es5, 所有evaluate的内容都需要先经过转码
class JsMode(Enum):
    ALL = 0
    JUST_ES5 = 1


JS_MODE = JsMode.JUST_ES5


# 注入js代码的cache, 无需重复读文件
class JsCache(dict):
    # 部分简单的代码直接定义
    common_code = {
        "clearMockAuth": "function() {global.__minium_auth_setting = {}}",  # 清除mock的auth信息
        "getAllPagesPath": "function(){return __wxConfig.pages}",  # 获取小程序所有页面路径
    }

    def __init__(self, *args, **kwargs):
        super(JsCache, self).__init__(*args, **kwargs)
        for mode in JsMode:
            self[mode] = copy.deepcopy(JsCache.common_code or {})

    def getJsCode(self, filename, mode: JsMode = None) -> str:
        if mode is None:
            mode = JS_MODE
        return self[mode][filename]

    def setJsCode(self, filename, content: str, mode: JsMode = None):
        if mode is None:
            mode = JS_MODE
        self[mode][filename] = content

    def has(self, filename, mode: JsMode = None):
        if mode is None:
            mode = JS_MODE
        return filename in self[mode]


JS_CACHE = JsCache()


def setInjectJsMode(mode: JsMode):
    global JS_MODE
    JS_MODE = mode


def getInjectJsCode(filename: str, format_info=None, mode: JsMode = None) -> str:
    """
    :param filename: {JS_PATH} 中 JS 文件名字（不需要后缀）
    :param format_info: JS 内容中需要进行格式化的信息, 如内容中包含 `%s` `%(arg)s` 等的可格式化信息

    部分注入的js需要动态格式化且%(\w+)s这类的语法会导致无法压缩, 使用 `r"$_\w+_$"` 替换JS中的 `%(\w+)s`, 在读取后替换回可格式化格式
    exp: %(origin)s 在js代码中使用 $_origin_$ 代替
    """
    if mode is None:
        mode = JS_MODE
    if JS_CACHE.has(filename, mode):
        if format_info:
            return JS_CACHE.getJsCode(filename, mode) % format_info
        return JS_CACHE.getJsCode(filename, mode)
    if mode is JsMode.JUST_ES5:
        min_js_file_path = os.path.join(ES5_JS_PATH, "%s.js" % filename)
    else:
        min_js_file_path = os.path.join(MIN_JS_PATH, "%s.js" % filename)
    is_min_file = False
    if os.path.isfile(min_js_file_path):
        js_file_path = min_js_file_path
        is_min_file = True
    else:
        if mode is JsMode.JUST_ES5:
            raise TypeError(f"can't find {filename}.js in just es5 mode")
        js_file_path = os.path.join(JS_PATH, "%s.js" % filename)  # source
    if os.path.isfile(js_file_path):
        with open(js_file_path, "r", encoding="utf8") as fd:
            content = fd.read()
            if not is_min_file:  # 没有压缩过的文件需要过滤一下注释
                content = re.sub(
                    r"^\s*//[^\n]*\s+", "", content, flags=re.M
                )  # 过滤文件头的注释
                content = re.sub(
                    r"^\s*/\*.*?\*/\s*", "", content, flags=re.S
                )  # 过滤文件头的注释
            content = re.sub(r"\$_(\w+)_\$", "%(\g<1>)s", content)  # 替换可格式化变量
            JS_CACHE.setJsCode(filename, content.strip(), mode)
            if format_info:
                return JS_CACHE.getJsCode(filename, mode) % format_info
            return JS_CACHE.getJsCode(filename, mode)
    raise RuntimeError("can't find {filename}.js, please check")


if __name__ == "__main__":
    print(getInjectJsCode("requestStack"))



#!/usr/bin/env python3
# Created by xiazeng on 2019-06-04
import sys
import inspect
import os.path
import re
import logging

from .minisuite import MiniSuite
from .exception import LoadCaseError
from . import minitest
import fnmatch

logger = logging.getLogger("minium")
target_base_class = minitest.AssertBase


def parse_document(doc_str):
    """
    解析用例的关键字段
    :param doc_str:
    :return:
    """
    pair = {}
    if not doc_str:
        return pair
    for line in doc_str.split("\n"):
        m = re.search(r"@(\w+)[\s:](.*)", line)
        if m:
            pair[m.group(1).lower()] = m.group(2)
    return pair


def import_module(module_name):
    module_name = module_name.strip()
    comps = module_name.split(".")
    mod = __import__(module_name)
    if len(comps) < 2:
        return mod
    else:
        for name in comps[1:]:
            mod = getattr(mod, name)
    return mod


def doc_format(doc_str):
    """
    去掉注释前面的空格
    :param doc_str:
    :return:
    """
    if doc_str is None:
        return None
    old_lines = doc_str.split("\n")
    min_space_count = 0
    for line in old_lines:
        line = line.replace("\t", "    ")
        space_count = 0
        for a in line:
            if a == " ":
                space_count += 1
            else:
                break
        if min_space_count < space_count:
            min_space_count = space_count
        if len(line) < min_space_count:
            min_space_count = len(line)
    new_lines = [line[min_space_count:] for line in old_lines]
    try:
        return "\n".join(new_lines).strip()
    except:
        return doc_str


def get_test_methods(test_case_class):
    s = []
    for name, value in inspect.getmembers(test_case_class):
        if name.startswith("test"):
            if callable(getattr(test_case_class, name)):
                for attr_name, func_obj in inspect.getmembers(value):
                    if attr_name == "__code__":
                        pair = parse_document(value.__doc__)
                        s.append(
                            {
                                "human_case_name": pair.get("name"),
                                "name": name,
                                "start": func_obj.co_firstlineno,
                                "filename": func_obj.co_filename,
                                "doc": doc_format(value.__doc__),
                                "co_names": func_obj.co_names,
                            }
                        )
    return s


def get_mod(module_name):
    try:
        mod = import_module(module_name)
        return mod
    except ModuleNotFoundError as e:
        logger.exception("ignore module: %s", module_name)
        return LoadCaseError(e)


def get_mod_members(mod):
    mod_memebers = {}
    for memname, memobj in inspect.getmembers(mod):
        if inspect.isclass(memobj) and issubclass(memobj, target_base_class):
            mod_memebers[memname] = memobj
    return mod_memebers


# 获取叶子结点
def get_leaves(mod_memebers):
    leaves = []
    has_subclass_set = set()
    mod_list = list(mod_memebers.values())
    for i in range(len(mod_list)):
        obj = mod_list[i]
        if obj in has_subclass_set:  # 继承的特性，如果这个节点还有父节点，肯定已经被其子节点检查出来
            continue
        if i == len(mod_list) - 1:
            leaves.append(obj)
            continue
        for j in range(i + 1, len(mod_list)):
            cmp_obj = mod_list[j]
            if issubclass(cmp_obj, obj):
                has_subclass_set.add(obj)
            elif issubclass(obj, cmp_obj):
                has_subclass_set.add(cmp_obj)
        if obj not in has_subclass_set:
            leaves.append(obj)
    return leaves


def load_module(path, pkg: str, mini_suite: MiniSuite = None):
    sys.path.append(path)
    test_cases = {}

    human_name_mapping = {}

    # 精确查找，现将 pkg 分段 xxx.yyy.zzz.case
    pkg_path = pkg.split(".")
    # 最后一个为 case 文件
    pkg_name = pkg_path[-1]
    # 前面的 module name 全部加到路径里面去, 但是 path 不能动，后面会用到
    search_path = path
    for d in pkg_path[:-1]:
        search_path = os.path.join(search_path, d)
    if not os.path.exists(search_path):
        raise Exception(f"path to 「{pkg}」 is not exists, please check it")
    for root, dirs, filenames in os.walk(search_path):
        for filename in filenames:
            if not filename.endswith(".py"):
                continue

            if not fnmatch.fnmatch(filename.split(".")[0], pkg_name):
                continue

            module_path = os.path.join(root, filename)
            module_path = os.path.relpath(
                os.path.abspath(module_path), os.path.abspath(path)
            )
            module_name = ".".join(
                module_path.replace(os.path.sep, ".").split(".")[:-1]
            )
            has_hit_names = set()
            mod = get_mod(module_name)
            if not mod or isinstance(mod, ModuleNotFoundError):
                if mini_suite:
                    mini_suite.set_pkg_fail(module_name, mod)
                continue

            human_name = getattr(mod, "NAME", None)
            human_id = getattr(mod, "ID", 0)
            if human_name:
                if module_name.endswith(".__init__"):
                    module_name = module_name[: -1 * len(".__init__")]
                human_name_mapping[module_name] = {
                    "human_case_module": human_name,
                    "module_id": human_id,
                }

            mod_memebers = get_mod_members(mod)
            leaves = get_leaves(mod_memebers)
            for leaf_obj in leaves:
                case_list = get_test_methods(leaf_obj)
                if case_list:
                    if module_name not in test_cases:
                        test_cases[module_name] = {
                            "class_name": leaf_obj.__name__,
                            "case_list": case_list,
                            "human_case_module": human_name,
                            "module_id": human_id,
                        }
                    else:
                        test_cases[module_name]["case_list"] += case_list
            if mini_suite:
                mini_suite.set_pkg_success(module_name)

    # 更新模块的可读性字段
    for module_name, module_info in test_cases.items():
        module_info.update(update_module_info(module_name, human_name_mapping))

    return test_cases


def update_module_info(module_name, human_name_mapping):
    module_sp = module_name.split(".")
    human_info = {}
    for i in range(len(module_sp)):
        m = ".".join(module_sp[: len(module_sp) - i])
        if m in human_name_mapping:
            human_info = human_name_mapping[m]
    return human_info


def find_test_class(module_name, test_name):
    for cls in get_test_classes(module_name):
        if test_name in dir(cls):
            return cls


def get_test_classes(module_name):
    classes = []
    test_module = import_module(module_name)
    for name in dir(test_module):
        obj = getattr(test_module, name)
        if isinstance(obj, type):
            if issubclass(obj, minitest.AssertBase):
                classes.append(obj)
    # 子类放前面, 基类放后面
    for i in range(len(classes)):
        if i == len(classes) - 1:
            break
        for j in range(i + 1, len(classes)):
            if issubclass(classes[j], classes[i]):
                tmp = classes[i]
                classes[i] = classes[j]
                classes[j] = tmp
    return classes



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Created by xiazeng on 2018/6/26

LauncherUI = "com.tencent.mm.ui.LauncherUI"

APP_LAUNCHER = "com.tencent.mm.plugin.appbrand.ui.AppBrandLauncherUI"
LUGGAGE_LAUNCHER = "com.tencent.wmpf.dev.MainActivity"
ALBUM = "com.tencent.mm.plugin.gallery.ui.AlbumPreviewUI"

BindContactUI = "com.tencent.mm.plugin.account.bind.ui.BindMContactUI"

SelectConversationUI = "com.tencent.mm.ui.transmit.SelectConversationUI"
SoSoProxyUI = "com.tencent.mm.plugin.location_soso.SoSoProxyUI"
AppInstallerUI = "com.tencent.mm.sandbox.updater.AppInstallerUI"


# 小程序预览图片
GestureGalleryUI = "com.tencent.mm.plugin.subapp.ui.gallery.GestureGalleryUI"
PAY_WalletPayUI = "com.tencent.mm.plugin.wallet.pay.ui.WalletPayUI"

# 小程序的界面， 有可能出现 com.tencent.mm.plugin.appbrand.ui.AppBrandUI1
AppBrandUI = "com.tencent.mm.plugin.appbrand.ui.AppBrandUI"

# 授权页面
AppBrandAuthorizeUI = "com.tencent.mm.plugin.appbrand.ui.AppBrandAuthorizeUI"
AppBrandAuthorizeUI_1 = "com.tencent.mm.plugin.appbrand.ui.AppBrandAuthorizeUI1"
AppBrandAuthorizeUI_2 = "com.tencent.mm.plugin.appbrand.ui.AppBrandAuthorizeUI2"
#地理位置页面
LocationUI = "com.tencent.mm.plugin.location_soso.SoSoProxyUI"
# 登录界面
LoginPasswordUI = "com.tencent.mm.plugin.account.ui.LoginPasswordUI"


AppBrandOpenWeRunSettingUI = "com.tencent.mm.plugin.appbrand.ui.AppBrandOpenWeRunSettingUI"




#!/usr/bin/env python3

"""
Author:         xiazeng
Create time:    2021/12/28
Description:    

"""
import json
import time
import unittest

from attestbase import AtTestBase
import at.core.uixml


class UiXMLTest(unittest.TestCase):
    def test_json_to_views(self):
        ui_views = self.at.java_driver.get_ui_views()
        node_json = ui_views[0].dump_to_jsons()
        for ui_view in at.core.uixml.json_2_ui_views(node_json):
            print(ui_view)

    def test_remove_kinda(self):
        content = open("/Users/xiazeng/Downloads/next_step.xml", 'r').read()
        print(type(content))
        remove_content = at.core.uixml.remove_kinda_window(content).decode("utf-8")
        print(remove_content)
        # open("/Users/xiazeng/Downloads/next_step_decode.xml", 'w').write(remove_content.decode("utf-8"))
        # content = open("/Users/xiazeng/Downloads/window_dump.xml", 'r').read().encode('utf-8')
        # remove_content = at.core.uixml.remove_kinda_window(content)
        # open("/Users/xiazeng/Downloads/window_dump_decode.xml", 'w').write(remove_content.decode('utf-8'))

    def test_xpath(self):

        # filename = '/Users/xiazeng/Downloads/1645709269125.xml'
        filename = '/Users/xiazeng/Downloads/1645775262986.xml'
        s = time.time()
        ui_views = at.core.uixml.window_dump_parse(filename)
        for ui_view in ui_views:
            print(ui_view)
            print(ui_view.get_xpath())
            print(ui_view.get_xpath(True))
            print(ui_view.get_xpath(True, True))
            print(ui_view.get_xpath(False, True))
            break
        index = 0
        # for ui_view in ui_views:
        #     for ot in ui_views:
        #         xpath = ot.get_xpath()
        #         index += 1
        print(index, time.time() - s)



from unittest import *
from .case import (
    addModuleCleanup,
    TestCase,
    FunctionTestCase,
    SkipTest,
    skip,
    skipIf,
    skipUnless,
    expectedFailure,
    expectedException,
)
from .suite import BaseTestSuite, TestSuite



#!/usr/bin/env python3
# Created by xiazeng on 2019-05-22
import sys
import os

lib_path = os.path.abspath(os.path.join(".."))
# print(lib_path)
sys.path.append(lib_path)

import requests
import wda
from minium.native.qq_native.qbasenative import QBaseNative, NativeError
from minium.native.lib.wx_wda import *
from urllib.parse import quote


class QiOSNative(QBaseNative):
    def __init__(self, json_conf):
        super(QiOSNative, self).__init__(json_conf)
        self.udid = json_conf.get("device_info").get("udid")  # get("device_desire").
        self.wda_project_path = json_conf.get("wda_project_path")  # get("device_desire").
        self.wda_runner = WebDriverRunner(self.udid, self.wda_project_path)
        self.app = None
        self.bundle_id = json_conf.get("bundle_id")

    ###############################
    #                    interface                        #
    ###############################

    # 打开scheme
    def open_remote_debug(self, scheme_url):
        # self.start_qq()
        # driver = wda.Client()
        # session = driver.session(bundle_id="com.tencent.qq.dailybuild")
        self.connect_QQminiApp(scheme_url)
        self.app.session(name=u"消息").wait_gone()
        time.sleep(20)

    def start_qq(self):
        """
        启动QQ
        :return:
        """
        device = DeviceTool(self.udid)
        device.find_app(bundle_id=self.bundle_id)  # com.tencent.mqq
        for i in range(3):
            try:
                logger.info("第 %d 次启动QQ, 共3次机会" % (i + 1))
                self.app = WdaUI(
                    server_url="http://localhost:%s" % self.wda_runner.port,
                    bundle_id=self.bundle_id,
                )
                self.app.session.set_alert_callback(self._alert_callback) if callable(
                    self._alert_callback
                ) else logger.error("Alert callback would not callable")
                logger.info("QQ启动成功")
                return
            except Exception as e:
                logger.error("setup error: 第 %d 次启动QQ失败: %s" % ((i + 1), str(e)))
                logger.info("正在重启 WebDriverAgent ...")
                self.wda_runner.start_driver()

    def start_app(self):
        self.start_qq()

    def stop_app(self):
        pass

    # 通过scheme链接启动手Q小程序
    def connect_QQminiApp(self, scheme_url):
        """
        :param scheme_url:
        :return:
        """
        scheme_url_urlencode = quote(scheme_url, "utf-8")
        current_path = str(os.path.abspath(os.path.dirname(__file__)))
        mini_app_command = (
            "ios-deploy  --justlaunch --bundle "
            + "%s/schemeAgent.app" % (current_path)
            + " -a "
            + "mqqapi://microapp/open?url=%s" % (scheme_url_urlencode)
        )
        os.system(mini_app_command)

    def screen_shot(self, filename: str, return_format: str = "raw") -> object:
        """ß
        截图
        :param filename: 文件存放的路径
        :param return_format: 除了将截图保存在本地之外, 需要返回的图片内容格式: raw(default) or pillow
        :return: raw data or PIL.Image
        """
        return self.app.client.screenshot(png_filename=filename, format=return_format)

    def pick_media_file(
        self,
        cap_type="camera",
        media_type="photo",
        original=False,
        duration=5.0,
        names=None,
    ):
        """
        获取媒体文件
        :param cap_type: camera: 拍摄 | album: 从相册获取
        :param names: 传入一个 list 选择多张照片或者视频(照片和视频不能同时选择, 且图片最多 9 张, 视频一次只能选择一个)
        :param media_type: photo 或者 video
        :param duration: 拍摄时长
        :param original: 是否选择原图(仅图片)
        :return:
        """
        if cap_type == "album" and names is None:
            Exception("从相册选择照片必须提供照片名称, 可以通过 wda inspector 查看照片名称")
        if cap_type == "camera":
            self._capture_photo(media_type=media_type, duration=duration)
        elif cap_type == "album":
            if media_type == "photo":
                if isinstance(names, str):
                    names = [names]
                self._select_photos_from_album(names=names, original=original)
            elif media_type == "video":
                if isinstance(names, list):
                    names = names[0]
                self._select_video_from_album(name=names)

    def input_text(self, text):
        """
        input 组件填写文字(使用此函数之前必须确保输入框处于输入状态)
        :param text: 内容
        :return:
        """
        self.app.session(className="TextField").set_text(text)

    def input_clear(self):
        """
        input 组件清除文字(使用此函数之前必须确保输入框处于输入状态)
        :return:
        """
        self.app.session(className="TextField").clear_text()

    def textarea_text(self, text: str, index=1):
        """
        给 textarea 输入文字需要在小程序层提供该 textarea 的 index 信息
        :param text: 内容
        :param index: 多个 textarea 同时存在一个页面从上往下排序, 计数从 1 开始
        :return:
        """
        self.app.session(className="TextView")[index].set_text(text)

    def textarea_clear(self, index=0):
        """
        给 textarea 清除文字需要在小程序层提供该 textarea 的 index 信息
        :param index: 多个 textarea 同时存在一个页面从上往下排序, 计数从 1 开始
        :return:
        """
        self.app.session(className="textView")[index].clear_text()

    def allow_login(self, answer=True):
        """
        处理app登陆确认弹框
        :return:
        """
        if answer:
            self.app.session(className="Button", name="Allow").get(timeout=10.0).click()
        else:
            # todo @locker
            pass

    def allow_get_user_info(self, answer=True):
        """
        处理获取用户信息确认弹框
        :param answer: True or False
        :return:
        """
        if self.app.session(className="StaticText", nameContains="获取你的昵称").exists:
            self.app.session(className="Button", name="允许").get(
                timeout=10.0
            ).click() if answer else self.app.session(className="Button", name="取消").get(
                timeout=10.0
            ).click()

    def allow_get_location(self, answer=True):
        """
        处理获取位置信息确认弹框
        :param answer: True or False
        :return:
        """
        if self.app.session(className="StaticText", name="获取你的位置信息").exists:
            self.app.session(className="Button", name="允许").get(
                timeout=10.0
            ).click() if answer else self.app.session(className="Button", name="拒绝").get(
                timeout=10.0
            ).click()

    def handle_modal(self, btn_text="确定", title: str = None):
        """
        处理模态弹窗
        :param title: 传入弹窗的 title 可以校验当前弹窗是否为预期弹窗
        :param btn_text: 根据传入的 文字 进行点击
        :return:
        """
        if title:
            assert self.app.session(nameContains=title).exists, "没有出现预期弹窗: %s" % title
        self.app.session(className="Button", name=btn_text).get(timeout=10.0).click()

    def handle_action_sheet(self, item):
        """
        处理上拉菜单
        :param item: 要选择的 item
        :return:
        """
        self.app.session(className="ScrollView").child(className="Button", name=item).get(
            timeout=10.0
        ).click()

    def forward_miniprogram(self, names, text: str = None, create_new_chat=True):
        """
        通过右上角更多菜单转发小程序
        ps: 好友太多会有性能问题
        :type text: 分享携带的内容
        :param names: 要分享的人
        :param create_new_chat: 是否创建群聊
        :return:
        """
        self.app.session(className="Button", name="更多").get(timeout=10.0).click()
        self.app.session(className="Button", name="转发").get(timeout=10.0).click()
        self.app.session(className="StaticText", name="创建新的聊天").get(timeout=10.0).click()
        if isinstance(names, str):
            names = [names]
        for name in names:
            self.app.session(className="TextField", name="搜索").get(timeout=10.0).set_text(name)
            # count = 10
            self.app.session(className="StaticText", name=name, visible=True).get(
                timeout=10.0
            ).click()
            # count -= 1
        self.app.session(className="Button", nameContains="完成").get(timeout=10.0).click()
        self.app.session(className="Button", name="发送").get(timeout=10.0).click()

    def forward_miniprogram_inside(self, names, create_new_chat=True):
        """
        小程序内触发转发小程序
        ps: 好友太多会有性能问题
        :param names: 要分享的人
        :param create_new_chat: 是否创建群聊
        :return:
        """
        self.app.session(className="StaticText", name="创建新的聊天").get(timeout=10.0).click()
        if isinstance(names, str):
            names = [names]
        for name in names:
            self.app.session(className="TextField", name="搜索").get(timeout=10.0).set_text(name)
            # count = 10
            self.app.session(className="StaticText", name=name, visible=True).get(
                timeout=10.0
            ).click()
            # count -= 1
        self.app.session(className="Button", nameContains="完成").get(timeout=10.0).click()
        self.app.session(className="Button", name="发送").get(timeout=10.0).click()

    def send_custom_message(self, message: str = None):
        """
        处理小程序 im 发送自定义消息
        :param message: 消息内容
        :return:
        """
        self.app.session(className="TextView").set_text(message + "\n")

    def phone_call(self):
        """
        处理小程序拨打电话
        :return:
        """
        self.app.session(className="Button", name="呼叫").get(timeout=10.0).click()
        self.app.session.alert.accept()

    def map_select_location(self, name: str):
        """
        原生地图组件选择位置
        :param name: 位置名称
        :return:
        """
        self.app.session(className="SearchField", name="搜索地点").get(timeout=10.0).set_text(name)
        self.app.session(name=name, className="StaticText").get(timeout=10.0).click()
        while self.app.session(name=name, className="StaticText").exists:
            try:
                self.app.session(name=name, className="StaticText").get(timeout=10.0).click()
                if self.app.session(className="Button", name="确定").exists:
                    break
            except Exception as e:
                logger.warning(str(e))

        self.app.session(className="Button", name="确定").get(timeout=10.0).click()

    def map_back_to_mp(self):
        """
        原生地图组件查看定位,返回小程序
        :return:
        """
        self.app.session(className="Button", name="返回").get(timeout=10.0).click()

    def deactivate(self, duration):
        """
        使app进入后台一段时间, 再切回前台
        :param duration: float
        :return: NULL
        """
        self.app.session.deactivate(duration=duration)

    @property
    def orientation(self):
        """
        获取屏幕方向
        :return:
        """
        return self.app.session.orientation()

    @orientation.setter
    def orientation(self, value):
        """
        设置屏幕方向
        :param value: (string) LANDSCAPE | PORTRAIT | UIA_DEVICE_ORIENTATION_LANDSCAPERIGHT |
                    UIA_DEVICE_ORIENTATION_PORTRAIT_UPSIDEDOWN
        :return:
        """
        self.app.session.orientation(value)

    def release(self):
        """
        remove port forward process
        :return:
        """
        self.wda_runner.remove_iproxy()

    ###############################
    #                      private                         #
    ###############################

    def _capture_photo(self, media_type, duration=10.0):
        """
        capture a photo by camera
        :param media_type: photo or video
        :param duration: capture duration
        :return:
        """
        if media_type == "photo":
            self.app.session(text="拍照").get(timeout=10.0).click()
            self.app.session(name="PhotoCapture").get(timeout=10.0).click()
        elif media_type == "video":
            self.app.session(text="拍摄").get(timeout=10.0).click()
            self.app.session(name="VideoCapture").get(timeout=10.0).click()
            time.sleep(duration)
            self.app.session(name="VideoCapture").get(timeout=10.0).click()
        time.sleep(2.0)
        while self.app.session(nameContains="Use ").exists:
            try:
                self.app.session(nameContains="Use ").get(timeout=10.0).click()
            except Exception as e:
                logger.warning(str(e))

    def _select_photos_from_album(self, names: list, original=False):
        """
        select photos from album
        :param names: photo name list
        :param original: use original photo or not
        :return:
        """
        self.app.session(text="从手机相册选择").get(timeout=10.0).click()
        # self.app.session(text="原图发送").get(timeout=10.0)
        for name in names:
            rect = self.app.session(nameContains=name).bounds
            self.app.session.click(rect.x + rect.width - 10, rect.y + 10)
        if original:
            self.app.session(text="原图发送").get(timeout=10.0).click()
        self.app.session(text="完成").get(timeout=10.0).click()
        self.app.session(text="原图发送").wait_gone(timeout=10.0)

    def _select_video_from_album(self, name: str):
        """
        select video from album
        :param name: video file name
        :return:
        """
        self.app.session(text="从手机相册选择").get(timeout=10.0).click()
        rect = self.app.session(text="发送").get(timeout=10.0).bounds
        self.app.session(nameContains=name).get(timeout=10.0).click()
        self.app.session(text="取消").wait_gone(timeout=10.0)
        self.app.session.click(rect.x + 10, rect.y + 10)
        self.app.session(text="发送").wait_gone(timeout=300.0)

    def get_authorize_settings(self):
        """
        todo @locker
        :return:
        """
        pass

    def back_from_authorize_setting(self):
        """
        todo @locker
        :return:
        """
        self.app.session(className="Button", name="返回").get(timeout=10.0).click()

    def authorize_page_checkbox_enable(self, name, enable):
        """
        todo @locker
        :return:
        """
        pass

    @staticmethod
    def _alert_callback(session):
        """
        auto accept when system alert view popup
        :return:
        """
        logger.info("出现弹框, 默认接受")
        session.alert.accept()


if __name__ == "__main__":

    scheme_url = text = quote("http://qq.com/ok", "utf-8")
    mini_app_command = (
        "ios-deploy  --justlaunch --bundle "
        + "./schemeAgent.app"
        + " -a "
        + "mqqapi://microapp/open?url=%s" % (scheme_url)
    )
    print(mini_app_command)
    with open("/Users/chargo/minium/test_yy/demo/config.json") as json_file:
        import json

        conf = json.load(json_file)
        nv = QiOSNative(conf.get("device_desire"))
        nv.start_qq()
        # input("xxxxxxxxxxxxxxxxxxxxx")
        # open miniprogram by scheme url
        # nv.connect_miniApp(miniapp_url)

        # test shotcut
        # nv.screen_shot('./result_img')

        # nv.pick_media_file(cap_type="album", media_type="video", names="照片23")
        # nv.input_text("hi, locker")
        # nv.input_clear()
        # nv.textarea_text(text="hi, locker",index=2)
        # nv.allow_login()
        # nv.allow_get_user_info(answer=False)
        # nv.handle_modal(title="弹窗标题", answer="确定")
        # nv.handle_action_sheet(item="item1")
        # nv.forward_miniprogram(["恒瑜_Sherlock"])
        # nv.send_custom_message("你好")
        # nv.phone_call()
        # nv.map_select_location("广州塔")
        # nv.map_back_to_mp()
        # for i in range(101):
        #     nv.handle_slider(i, index=2)
        # nv.handle_slider(0.0, index=2)
        # nv.handle_switch(0)
        # nv.remove_iproxy()



# -*- coding: utf-8 -*-
# Created by xiazeng on 2020/3/10
# refer: https://github.com/openstf/minitouch
import os.path
import select
import logging
import os
import sys
import re
import socket
import time
from at.core.config import STF_LIB_PATH
from at.core.adbwrap import AdbWrap
from at.utils.commonhelper import pick_unuse_port, get_std_encoding
from at.utils.nbsp import NonBlockingStreamReader

logger = logging.getLogger()


class MiniTouch:
    sock = None  # type: socket.socket

    def __init__(self, serial=None, input_event=None, display_id=None):
        self.input_event = input_event
        self.display_id = display_id
        self.adb = AdbWrap.apply_adb(serial)
        self.path_in_android = "/data/local/tmp/minitouch"
        self.pc_port = pick_unuse_port()
        self.socket_name = "minitouch_%s" % self.pc_port
        self.max_x, self.max_y = None, None
        self.sock = None
        self._check_push()
        self._start_server()
        self._start_client()
        self.nbsp = None

    def _check_push(self):
        abi = self.adb.get_property("ro.product.cpu.abi")
        path = os.path.join(STF_LIB_PATH, abi, "minitouch").replace("\\", r"\\")
        if self.adb.compare_push(path, self.path_in_android):
            self.adb.run_shell("chmod 755 %s" % self.path_in_android)

    def _start_server(self):
        self.adb.forward_name("tcp:%s" % self.pc_port, "localabstract:"+self.socket_name)
        if self.input_event:
            p = self.adb.run_shell(
                "/data/local/tmp/minitouch -n '{0}' -d '{1}' 2>&1".format(self.socket_name, self.input_event), False)
        else:
            p = self.adb.run_shell("/data/local/tmp/minitouch -n '{0}' 2>&1".format(self.socket_name), False)
        nbsp = NonBlockingStreamReader(p.stdout, name="minitouch_output")
        timeout = 10
        s = time.time()
        while True:
            line = nbsp.readline(timeout=5.0)
            logger.debug(line)
            if line is None:
                raise RuntimeError("minitouch setup timeout")

            line = line.decode(get_std_encoding(sys.stdout))

            # 识别出setup成功的log，并匹配出max_x, max_y
            m = re.match("Type \w touch device .+ \((\d+)x(\d+) with \d+ contacts\) detected on .+ \(.+\)", line)
            if m:
                self.max_x, self.max_y = int(m.group(1)), int(m.group(2))
                break
            else:
                logging.info("not match")
            if time.time() - s > timeout:
                raise RuntimeError("minitouch setup timeout:%dms" % ((time.time() - s)*1000, ))
        self.nbsp = nbsp
        self.server_proc = p
        if self.server_proc.poll() is not None:
            raise RuntimeError("minitouch not running")
        # reg_cleanup(self.server_proc.kill)
        return p

    def _start_client(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect(("127.0.0.1", self.pc_port))
        sock.settimeout(1)
        ready = select.select([sock], [], [], 2)
        header = b""
        if ready:
            while True:
                try:
                    header += sock.recv(4096)  # size is not strict, so use raw socket.recv
                except socket.timeout:
                    # raise RuntimeError("minitouch setup client error")
                    logger.warning("minitouch header not recved")
                    break
                if header.count(b'\n') >= 3:
                    break
        else:
            raise RuntimeError("connect timeout ")
        logger.debug("minitouch header:%s", repr(header))
        self.sock = sock

    def release(self):
        if self.nbsp:
            self.nbsp.kill()
        s = time.time()
        self.adb.forward_remove(self.pc_port)
        if not self.server_proc:
            return
        self.server_proc.kill()
        if self.server_proc.poll() is not None:
            self.adb.kill_pid(self.server_proc.pid)

    def ensure_running(self):
        if self.server_proc.poll() is not None:
            raise RuntimeError("minitouch not running")
        if not self.sock:
            raise RuntimeError("client not connected")

    def send(self, data):
        logger.debug(data)
        self.sock.send(data)

    def touch_down(self, x, y, contract_id=0):
        self.ensure_running()
        cmd = b'd %d %d %d 50\nc\n' % (contract_id, x, y)
        self.send(cmd)

    def touch_move(self, x, y, contract_id=0):
        self.ensure_running()
        cmd = b'm %d %d %d 50\nc\n' % (contract_id, x, y)
        self.send(cmd)

    def touch_up(self, contract_id=0):
        self.ensure_running()
        cmd = b'u %d\nc\n' % contract_id
        self.send(cmd)

    def click(self, x, y, sleep_time=0.01):
        self.ensure_running()
        self.touch_down(x, y)
        time.sleep(sleep_time)
        self.touch_up()
        time.sleep(0.5)

    def long_click(self, x, y, sleep_time=3):
        self.click(x, y, sleep_time)

    def double_click(self, x, y):
        self.click(x, y)
        time.sleep(0.05)
        self.click(x, y)

    def swipe(self, x, y, dst_x, dst_y, steps=20, time_costs=0.5):
        self.touch_down(x, y)
        delta_x = (dst_x-x)/steps
        delta_y = (dst_y-y)/steps
        delta_time = time_costs / steps
        for i in range(steps-1):
            self.touch_move(x+delta_x*(i+1), y + delta_y*(y+1))
            time.sleep(delta_time)
        self.touch_move(dst_x, dst_y)
        self.touch_up()


